# torchcell/datasets/scerevisiae/__init__.py
# [[torchcell.datasets.scerevisiae.__init__]]
# https://github.com/Mjvolk3/torchcell/tree/main/torchcell/datasets/scerevisiae/__init__.py
# Test file: tests/torchcell/datasets/scerevisiae/test___init__.py

from .costanzo2016 import DmfCostanzo2016Dataset, SmfCostanzo2016Dataset
from .kuzmin2018 import SmfKuzmin2018Dataset, DmfKuzmin2018Dataset, TmfKuzmin2018Dataset

# from .baryshnikovna2010 import Baryshnikovna2010Dataset

costanzo_datasets = ["SmfCostanzo2016Dataset", "DmfCostanzo2016Dataset"]

kuzmin_datasets = [
    "SmfKuzmin2018Dataset",
    "DmfKuzmin2018Dataset",
    "TmfKuzmin2018Dataset",
]

__all__ = costanzo_datasets + kuzmin_datasets



# torchcell/models/deep_set_transformer.py
# [[torchcell.models.deep_set_transformer]]
# https://github.com/Mjvolk3/torchcell/tree/main/torchcell/models/deep_set_transformer.py
# Test file: torchcell/models/test_deep_set_transformer.py

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch_scatter import scatter_add


class SelfAttention(nn.Module):
    def __init__(self, dim_in: int, dim_out: int, num_heads: int = 1):
        super().__init__()
        self.query = nn.Linear(dim_in, dim_out * num_heads)
        self.key = nn.Linear(dim_in, dim_out * num_heads)
        self.value = nn.Linear(dim_in, dim_out * num_heads)
        self.dim_out = dim_out
        self.num_heads = num_heads

    def forward(self, x):
        Q = self.query(x)
        K = self.key(x)
        V = self.value(x)

        # Reshaping to have the num_heads represented in the tensor
        Q = Q.view(Q.size(0), -1, self.num_heads, self.dim_out).permute(2, 0, 1, 3)
        K = K.view(K.size(0), -1, self.num_heads, self.dim_out).permute(2, 0, 1, 3)
        V = V.view(V.size(0), -1, self.num_heads, self.dim_out).permute(2, 0, 1, 3)

        attn = F.softmax(Q @ K.transpose(-2, -1) / self.dim_out**0.5, dim=-1)
        out = attn @ V
        out = out.permute(1, 2, 0, 3).contiguous()

        return (out, attn)


class DeepSetTransformer(nn.Module):
    def __init__(
        self,
        input_dim: int,
        node_layers: list[int],
        set_layers: list[int],
        dropout_prob: float = 0.2,
        norm: str = "batch",
        activation: str = "relu",
        skip_node: bool = False,
        skip_set: bool = False,
        num_heads: int = 1,
        is_concat_attention: bool = True,
    ):
        super().__init__()

        assert norm in ["batch", "instance", "layer"], "Invalid norm type"
        activations = {
            "relu": nn.ReLU(),
            "gelu": nn.GELU(),
            "leaky_relu": nn.LeakyReLU(),
            "sigmoid": nn.Sigmoid(),
        }
        assert activation in activations.keys(), "Invalid activation type"

        self.skip_node = skip_node
        self.skip_set = skip_set
        self.is_concat_attention = is_concat_attention

        def create_block(in_dim, out_dim, norm, activation):
            block = [nn.Linear(in_dim, out_dim)]
            if norm == "batch":
                block.append(nn.BatchNorm1d(out_dim))
            elif norm == "instance":
                block.append(nn.InstanceNorm1d(out_dim, affine=True))
            elif norm == "layer":
                block.append(nn.LayerNorm(out_dim))
            block.append(activations[activation])
            return nn.Sequential(*block)

        in_dim = input_dim
        node_modules = []
        for out_dim in node_layers:
            node_modules.append(create_block(in_dim, out_dim, norm, activation))
            in_dim = out_dim
        self.node_layers = nn.ModuleList(node_modules)

        set_modules = []
        for i, out_dim in enumerate(set_layers):
            if i == 0 and self.is_concat_attention:
                in_dim = in_dim * num_heads
            set_modules.append(create_block(in_dim, out_dim, norm, activation))
            in_dim = out_dim
        set_modules.append(nn.Dropout(dropout_prob))
        self.set_layers = nn.ModuleList(set_modules)

        # Create Set Transformer layer between node_layers and set_layers
        self.set_transformer = SelfAttention(
            dim_in=node_layers[-1], dim_out=node_layers[-1], num_heads=num_heads
        )  # Here, using the same dimension for input and output and 1 head.

    def forward(self, x, batch):
        x_node = x
        for i, layer in enumerate(self.node_layers):
            out_node = layer(x_node)
            if self.skip_node and x_node.shape[-1] == out_node.shape[-1]:
                out_node += x_node  # Skip connection
            x_node = out_node

        # Apply Set Transformer Layer here
        x_transformed, attn_weights = self.set_transformer(x_node)
        if self.is_concat_attention:
            x_transformed_cast = x_transformed.reshape(x_transformed.shape[0], 1, 1, -1)
        else:
            x_transformed_cast = x_transformed.mean(dim=2, keepdim=True)

        x_summed = scatter_add(x_transformed_cast, batch, dim=0)

        x_set = x_summed
        for i, layer in enumerate(self.set_layers):
            out_set = layer(x_set)
            if self.skip_set and x_set.shape[-1] == out_set.shape[-1]:
                out_set += x_set  # Skip connection
            x_set = out_set

        x_set = x_set.squeeze(1, 2)

        if x_set.dim() == 1:
            x_set = x_set.unsqueeze(0)

        return (
            x_node,
            x_set,
            x_transformed.squeeze(1),
            attn_weights,
        )  # returning x_transformed as well, in case you need it.


def main():
    input_dim = 10
    node_layers = [64, 32, 32, 32]
    set_layers = [16, 8, 8]

    model = DeepSetTransformer(
        input_dim,
        node_layers,
        set_layers,
        norm="layer",
        activation="gelu",
        skip_node=True,
        skip_set=True,
        num_heads=3,
        is_concat_attention=False,
    )

    x = torch.rand(100, input_dim)
    batch = torch.cat([torch.full((20,), i, dtype=torch.long) for i in range(5)])
    print("x", x.shape)
    x_nodes, x_set, x_transformed, attn_weights = model(x, batch)

    print("x_set.shape", x_set.shape)
    print("x_nodes.shape", x_nodes.shape)
    print("x_transformed.shape", x_transformed.shape)
    print("attn_weights.shape", attn_weights.shape)


if __name__ == "__main__":
    main()



import logging
import os
import os.path as osp

import hydra
import lightning as L
import torch
from dotenv import load_dotenv
from lightning.pytorch.callbacks import ModelCheckpoint
from lightning.pytorch.loggers import WandbLogger
from omegaconf import DictConfig, OmegaConf
from torch_geometric.loader import DataLoader
from tqdm import tqdm

import wandb
from torchcell.datamodules import CellDataModule
from torchcell.datasets import (
    CellDataset,
    FungalUtrTransformerDataset,
    NucleotideTransformerDataset,
)
from torchcell.datasets.scerevisiae import DmfCostanzo2016Dataset
from torchcell.models import DeepSet
from torchcell.sequence.genome.scerevisiae.s288c import SCerevisiaeGenome
from torchcell.trainers import RegressionTask


def main():
    log = logging.getLogger(__name__)
    load_dotenv()
    DATA_ROOT = os.getenv("DATA_ROOT")

    # Get reference genome
    genome = SCerevisiaeGenome(data_root=osp.join(DATA_ROOT, "data/sgd/genome"))
    genome.drop_chrmt()
    genome.drop_empty_go()
    genome_gene_set = genome.gene_set

    # Sequence transformers
    nt_dataset = NucleotideTransformerDataset(
        root=osp.join(DATA_ROOT, "data/scerevisiae/nucleotide_transformer_embed"),
        genome=genome,
        transformer_model_name="nt_window_5979",
    )

    # Experiments
    experiments = DmfCostanzo2016Dataset(
        preprocess="low_dmf_std",
        root=osp.join(DATA_ROOT, "data/scerevisiae/costanzo2016"),
    )
    print(len(experiments))

    # Gather into CellDatset
    cell_dataset = CellDataset(
        root=osp.join(osp.join(DATA_ROOT, "data/scerevisiae/cell")),
        genome_gene_set=genome_gene_set,
        seq_embeddings=nt_dataset,
        experiments=experiments,
    )
    print(len(cell_dataset))
    data_loader = DataLoader(cell_dataset, batch_size=32, shuffle=True, num_workers=2)
    print(cell_dataset)
    print(cell_dataset[0])

    for batch in tqdm(data_loader):
        # print(f"Accessing LMDB from process {os.getpid()}")
        pass


if __name__ == "__main__":
    main()



# torchcell/datasets/scerevisiae/nucleotide_transformer.py
# [[torchcell.datasets.scerevisiae.nucleotide_transformer]]
# https://github.com/Mjvolk3/torchcell/tree/main/torchcell/datasets/scerevisiae/nucleotide_transformer.py
# Test file: tests/torchcell/datasets/scerevisiae/test_nucleotide_transformer.py

import os
from collections.abc import Callable
from typing import Optional

import torch
from torch_geometric.data import Data, InMemoryDataset
from tqdm import tqdm

from torchcell.datasets.embedding import BaseEmbeddingDataset
from torchcell.datasets.fungal_utr_transformer import FungalUtrTransformerDataset
from torchcell.datasets.nucleotide_transformer import NucleotideTransformerDataset
from torchcell.models.nucleotide_transformer import NucleotideTransformer
from torchcell.sequence.genome.scerevisiae.s288c import SCerevisiaeGenome

if __name__ == "__main__":
    # genome
    genome = SCerevisiaeGenome()
    # nucleotide transformer
    nucleotide_transformer_name = "nt_window_3utr_300_undersize"
    nt_dataset = NucleotideTransformerDataset(
        root="data/scerevisiae/nucleotide_transformer_embed",
        genome=genome,
        transformer_model_name=nucleotide_transformer_name,
    )
    # fungal utr transformer
    fungal_utr_transformer_name = "fut_window_3utr_300_undersize"
    fut_dataset = FungalUtrTransformerDataset(
        root="data/scerevisiae/fungal_utr_embed",
        genome=genome,
        transformer_model_name=fungal_utr_transformer_name,
    )
    # Combine the datasets
    print()



# %%
# Relies on env-intermine (python 3.5)
import json
import os
import os.path as osp

import pandas as pd
from dask import compute, delayed
from intermine.webservice import Service
from tqdm import tqdm

from torchcell.sc_graph import get_gene_list


def get_regulators(gene: str = "YOR202W") -> int:
    # Gene Regulation
    # Retrieve genes that are regulators of a given target gene.
    # Gene(target) -> Gene(Regulators).
    # "YOR202W" == "HIS3"
    service = Service("https://yeastmine.yeastgenome.org/yeastmine/service")
    template = service.get_template("GeneTarget_GeneFactor")
    rows = template.rows(
        A={"op": "LOOKUP", "value": gene, "extra_value": "S. cerevisiae"}
    )
    data_dict_list = []
    reg_dict = {}
    for row in rows:
        data_dict = {}
        data_dict["regulatoryRegions.regulator.symbol"] = row[
            "regulatoryRegions.regulator.symbol"
        ]
        data_dict["regulatoryRegions.regulator.secondaryIdentifier"] = row[
            "regulatoryRegions.regulator.secondaryIdentifier"
        ]
        data_dict["symbol"] = row["symbol"]
        data_dict["secondaryIdentifier"] = row["secondaryIdentifier"]
        data_dict["regulatoryRegions.regEvidence.ontologyTerm.name"] = row[
            "regulatoryRegions.regEvidence.ontologyTerm.name"
        ]
        data_dict["regulatoryRegions.regEvidence.ontologyTerm.identifier"] = row[
            "regulatoryRegions.regEvidence.ontologyTerm.identifier"
        ]
        data_dict["regulatoryRegions.experimentCondition"] = row[
            "regulatoryRegions.experimentCondition"
        ]
        data_dict["regulatoryRegions.strainBackground"] = row[
            "regulatoryRegions.strainBackground"
        ]
        data_dict["regulatoryRegions.regulationDirection"] = row[
            "regulatoryRegions.regulationDirection"
        ]
        data_dict["regulatoryRegions.publications.pubMedId"] = row[
            "regulatoryRegions.publications.pubMedId"
        ]
        data_dict["regulatoryRegions.regulatorType"] = row[
            "regulatoryRegions.regulatorType"
        ]
        data_dict["regulatoryRegions.datasource"] = row["regulatoryRegions.datasource"]
        data_dict["regulatoryRegions.annotationType"] = row[
            "regulatoryRegions.annotationType"
        ]
        data_dict_list.append(data_dict)
    reg_dict["regulators"] = data_dict_list
    # if reg_dict is empty set to None
    if reg_dict == {}:
        reg_dict["regulators"] = None
    return reg_dict


def get_physical_interactions(gene: str = "YFL039C") -> dict:
    # act1  == YFL039C
    # Physical Interactions
    # Retrieve all physical interactions for a specific gene. Gene(target) -> Physical Interactions.
    service = Service("https://yeastmine.yeastgenome.org/yeastmine/service")
    template = service.get_template("Gene_Physical_Interaction")
    rows = template.rows(A={"op": "LOOKUP", "value": gene, "extra_value": ""})

    data_dict_list = []
    physical_int_dict = {}
    for row in rows:
        data_dict = {}
        data_dict["primaryIdentifier"] = row["primaryIdentifier"]
        data_dict["symbol"] = row["symbol"]
        data_dict["secondaryIdentifier"] = row["secondaryIdentifier"]
        data_dict["sgdAlias"] = row["sgdAlias"]
        data_dict["name"] = row["name"]
        data_dict["organism.shortName"] = row["organism.shortName"]
        data_dict["interactions.details.annotationType"] = row[
            "interactions.details.annotationType"
        ]
        data_dict["interactions.details.role1"] = row["interactions.details.role1"]
        data_dict["interactions.participant2.symbol"] = row[
            "interactions.participant2.symbol"
        ]
        data_dict["interactions.participant2.secondaryIdentifier"] = row[
            "interactions.participant2.secondaryIdentifier"
        ]
        data_dict[
            "interactions.details.experiment.interactionDetectionMethods.identifier"
        ] = row[
            "interactions.details.experiment.interactionDetectionMethods.identifier"
        ]
        data_dict["interactions.details.experiment.name"] = row[
            "interactions.details.experiment.name"
        ]
        data_dict["interactions.details.relationshipType"] = row[
            "interactions.details.relationshipType"
        ]
        data_dict["interactions.details.note"] = row["interactions.details.note"]
        data_dict_list.append(data_dict)
    physical_int_dict["physical_interactions"] = data_dict_list
    # if physical_int_dict is empty set to None
    if physical_int_dict == {}:
        physical_int_dict["physical_interactions"] = None
    return physical_int_dict


def get_gene_interactions(gene: str = "YDR210W") -> dict:
    # Gene Interactions
    # Retrieve all gene interactions for a specific gene. Gene(target) -> Gene Interactions.
    service = Service("https://yeastmine.yeastgenome.org/yeastmine/service")
    template = service.get_template("Gene_Genetic_Interaction")
    rows = template.rows(A={"op": "LOOKUP", "value": gene, "extra_value": ""})

    data_dict_list = []
    gene_int_dict = {}
    for i, row in enumerate(rows):
        data_dict = {}
        data_dict["primaryIdentifier"] = row["primaryIdentifier"]
        data_dict["symbol"] = row["symbol"]
        data_dict["secondaryIdentifier"] = row["secondaryIdentifier"]
        data_dict["sgdAlias"] = row["sgdAlias"]
        data_dict["name"] = row["name"]
        data_dict["organism.shortName"] = row["organism.shortName"]
        data_dict["interactions.details.annotationType"] = row[
            "interactions.details.annotationType"
        ]
        data_dict["interactions.details.phenotype"] = row[
            "interactions.details.phenotype"
        ]
        # TODO - What is interactions.details.role1 describing? Looks ok for now... Can check after
        data_dict["interactions.details.role1"] = row["interactions.details.role1"]
        data_dict["interactions.participant2.symbol"] = row[
            "interactions.participant2.symbol"
        ]
        data_dict["interactions.participant2.secondaryIdentifier"] = row[
            "interactions.participant2.secondaryIdentifier"
        ]
        data_dict[
            "interactions.details.experiment.interactionDetectionMethods.identifier"
        ] = row[
            "interactions.details.experiment.interactionDetectionMethods.identifier"
        ]
        data_dict["interactions.details.experiment.name"] = row[
            "interactions.details.experiment.name"
        ]
        data_dict["interactions.details.relationshipType"] = row[
            "interactions.details.relationshipType"
        ]
        data_dict["interactions.details.note"] = row["interactions.details.note"]
        data_dict["interactions.alleleinteractions.allele1.name"] = row[
            "interactions.alleleinteractions.allele1.name"
        ]
        data_dict["interactions.alleleinteractions.allele2.name"] = row[
            "interactions.alleleinteractions.allele2.name"
        ]
        data_dict["interactions.alleleinteractions.pvalue"] = row[
            "interactions.alleleinteractions.pvalue"
        ]
        data_dict["interactions.alleleinteractions.sgaScore"] = row[
            "interactions.alleleinteractions.sgaScore"
        ]
        data_dict_list.append(data_dict)
    gene_int_dict["gene_interactions"] = data_dict_list
    # if gene_int_dict is empty set to None
    if gene_int_dict == {}:
        gene_int_dict["gene_interactions"] = None
    return gene_int_dict


def get_protein_abundance(gene: str = "tfc3") -> dict:
    # Retrieve protein abundance for the protein(s) encoded by a gene
    # Gene(target) -> Protein Abundance Experiments.
    service = Service("https://yeastmine.yeastgenome.org/yeastmine/service")
    template = service.get_template("Gene_ProteinAbundance")
    rows = template.rows(A={"op": "LOOKUP", "value": gene, "extra_value": ""})
    data_dict_list = []
    protein_abundance_dict = {}
    for row in rows:
        data_dict = {}
        data_dict["primaryIdentifier"] = row["primaryIdentifier"]
        data_dict["proteins.symbol"] = row["proteins.symbol"]
        data_dict["secondaryIdentifier"] = row["secondaryIdentifier"]
        data_dict["qualifier"] = row["qualifier"]
        data_dict["proteins.proteinAbundance.abundance"] = row[
            "proteins.proteinAbundance.abundance"
        ]
        data_dict["proteins.proteinAbundance.units"] = row[
            "proteins.proteinAbundance.units"
        ]
        data_dict["proteins.proteinAbundance.media"] = row[
            "proteins.proteinAbundance.media"
        ]
        data_dict["proteins.proteinAbundance.treatment"] = row[
            "proteins.proteinAbundance.treatment"
        ]
        data_dict["proteins.proteinAbundance.treatmentTime"] = row[
            "proteins.proteinAbundance.treatmentTime"
        ]
        data_dict["proteins.proteinAbundance.foldChange"] = row[
            "proteins.proteinAbundance.foldChange"
        ]
        data_dict["proteins.proteinAbundance.visualization"] = row[
            "proteins.proteinAbundance.visualization"
        ]
        data_dict["proteins.proteinAbundance.strainBackground"] = row[
            "proteins.proteinAbundance.strainBackground"
        ]
        data_dict["proteins.proteinAbundance.origPublication.pubMedId"] = row[
            "proteins.proteinAbundance.origPublication.pubMedId"
        ]
        data_dict["proteins.proteinAbundance.publication.pubMedId"] = row[
            "proteins.proteinAbundance.publication.pubMedId"
        ]
        data_dict_list.append(data_dict)
    protein_abundance_dict["protein_abundance"] = data_dict_list
    # if protein_abundance_dict is empty set to None
    if protein_abundance_dict == {}:
        protein_abundance_dict["protein_abundance"] = None
    return protein_abundance_dict


def get_median_protein_abundance(gene: str = "YAL001C") -> dict:
    # Gene median protein abundance
    # Retrieve all protein median abundance for a specific gene. Gene(target) -> median protein abundance.
    service = Service("https://yeastmine.yeastgenome.org/yeastmine/service")
    template = service.get_template("Protein_Median_Abundance")
    rows = template.rows(A={"op": "LOOKUP", "value": gene, "extra_value": ""})
    data_dict_list = []
    median_protein_abundance_dict = {}
    for row in rows:
        data_dict = {}
        data_dict["primaryIdentifier"] = row["primaryIdentifier"]
        data_dict["proteins.symbol"] = row["proteins.symbol"]
        data_dict["secondaryIdentifier"] = row["secondaryIdentifier"]
        data_dict["qualifier"] = row["qualifier"]
        data_dict["proteins.median"] = row["proteins.median"]
        data_dict["proteins.units"] = row["proteins.units"]
        data_dict["proteins.MAD"] = row["proteins.MAD"]
        data_dict["proteins.proteinAbundance.publication.pubMedId"] = row[
            "proteins.proteinAbundance.publication.pubMedId"
        ]
        data_dict_list.append(data_dict)
    median_protein_abundance_dict["median_protein_abundance"] = data_dict_list
    # if median_protein_abundance_dict is empty set to None
    if median_protein_abundance_dict == {}:
        median_protein_abundance_dict["median_protein_abundance"] = None
    return median_protein_abundance_dict


def get_protein_sequence(gene: str = "rad54") -> dict:
    # Gene(target) -> Gene protein sequence
    service = Service("https://yeastmine.yeastgenome.org/yeastmine/service")
    template = service.get_template("Gene_ProteinSequence")
    rows = template.rows(
        B={"op": "LOOKUP", "value": gene, "extra_value": "S. cerevisiae"}
    )
    data_dict_list = []
    protein_sequence_dict = {}
    for row in rows:
        data_dict = {}
        data_dict["secondaryIdentifier"] = row["secondaryIdentifier"]
        data_dict["symbol"] = row["symbol"]
        data_dict["name"] = row["name"]
        data_dict["sgdAlias"] = row["sgdAlias"]
        data_dict["proteins.sequence.residues"] = row["proteins.sequence.residues"]
        data_dict["proteins.sequence.length"] = row["proteins.sequence.length"]
        data_dict["featureType"] = row["featureType"]
        data_dict["proteins.symbol"] = row["proteins.symbol"]
        data_dict_list.append(data_dict)
    protein_sequence_dict["protein_sequence"] = data_dict_list
    # if protein_sequence_dict is empty set to None
    if protein_sequence_dict == {}:
        protein_sequence_dict["protein_sequence"] = None
    return protein_sequence_dict


def get_gene_sequence(gene: str = "rad54") -> dict:
    # Retrieve genomic DNA (DNA sequence with introns)
    # # Gene(target) -> Gene sequence
    service = Service("https://yeastmine.yeastgenome.org/yeastmine/service")
    template = service.get_template("Gene_GenomicDNA")
    rows = template.rows(
        E={"op": "LOOKUP", "value": gene, "extra_value": "S. cerevisiae"}
    )
    data_dict_list = []
    gene_sequence_dict = {}
    for row in rows:
        data_dict = {}
        data_dict["primaryIdentifier"] = row["primaryIdentifier"]
        data_dict["secondaryIdentifier"] = row["secondaryIdentifier"]
        data_dict["symbol"] = row["symbol"]
        data_dict["name"] = row["name"]
        data_dict["sgdAlias"] = row["sgdAlias"]
        data_dict["organism.shortName"] = row["organism.shortName"]
        data_dict["sequence.length"] = row["sequence.length"]
        data_dict["sequence.residues"] = row["sequence.residues"]
        data_dict["description"] = row["description"]
        data_dict["qualifier"] = row["qualifier"]
        data_dict_list.append(data_dict)
    gene_sequence_dict["gene_sequence"] = data_dict_list
    # if gene_sequence_dict is empty set to None
    if gene_sequence_dict == {}:
        gene_sequence_dict["gene_sequence"] = None
    return gene_sequence_dict


def get_go(gene: str = "YAL018C") -> dict:
    # Gene(target) -> GO terms
    service = Service("https://yeastmine.yeastgenome.org/yeastmine/service")
    template = service.get_template("Gene_GO")
    data_dict_list = []
    go_dict = {}
    rows = template.rows(A={"op": "LOOKUP", "value": gene, "extra_value": ""})
    for row in rows:
        data_dict = {}
        data_dict["primaryIdentifier"] = row["primaryIdentifier"]
        data_dict["secondaryIdentifier"] = row["secondaryIdentifier"]
        data_dict["symbol"] = row["symbol"]
        data_dict["featureType"] = row["featureType"]
        data_dict["qualifier"] = row["qualifier"]
        data_dict["goAnnotation.ontologyTerm.identifier"] = row[
            "goAnnotation.ontologyTerm.identifier"
        ]
        data_dict["goAnnotation.ontologyTerm.name"] = row[
            "goAnnotation.ontologyTerm.name"
        ]
        data_dict["goAnnotation.ontologyTerm.namespace"] = row[
            "goAnnotation.ontologyTerm.namespace"
        ]
        data_dict["goAnnotation.evidence.code.code"] = row[
            "goAnnotation.evidence.code.code"
        ]
        data_dict["goAnnotation.qualifier"] = row["goAnnotation.qualifier"]
        data_dict["goAnnotation.evidence.code.withText"] = row[
            "goAnnotation.evidence.code.withText"
        ]
        data_dict["goAnnotation.annotationExtension"] = row[
            "goAnnotation.annotationExtension"
        ]
        data_dict["goAnnotation.evidence.code.annotType"] = row[
            "goAnnotation.evidence.code.annotType"
        ]
        data_dict["goAnnotation.evidence.publications.pubMedId"] = row[
            "goAnnotation.evidence.publications.pubMedId"
        ]
        data_dict["goAnnotation.evidence.publications.citation"] = row[
            "goAnnotation.evidence.publications.citation"
        ]
        data_dict_list.append(data_dict)
    go_dict["go"] = data_dict_list
    # if go_dict is empty set to None
    if go_dict == {}:
        go_dict["go"] = None
    return go_dict


def get_protein_half_life(gene: str = "act1") -> dict:
    # Gene(target) -> protein half life
    # Retrieve Protein half-life for a given gene(s). This study had reported value
    # of >=100 hours for the most stable proteins. To facilitate sorting and
    # searching this values have been substituted  '10000' hours.

    service = Service("https://yeastmine.yeastgenome.org/yeastmine/service")
    template = service.get_template("Gene_Protein_Half_Life")
    rows = template.rows(
        A={"op": "LOOKUP", "value": gene, "extra_value": "S. cerevisiae"}
    )
    data_dict_list = []
    protein_half_life_dict = {}
    for row in rows:
        data_dict = {}
        data_dict["primaryIdentifier"] = row["primaryIdentifier"]
        data_dict["secondaryIdentifier"] = row["secondaryIdentifier"]
        data_dict["symbol"] = row["symbol"]
        data_dict["description"] = row["description"]
        data_dict["proteins.symbol"] = row["proteins.symbol"]
        data_dict["proteins.proteinHalfLife.experiment"] = row[
            "proteins.proteinHalfLife.experiment"
        ]
        data_dict["proteins.proteinHalfLife.value"] = row[
            "proteins.proteinHalfLife.value"
        ]
        data_dict["proteins.proteinHalfLife.units"] = row[
            "proteins.proteinHalfLife.units"
        ]
        data_dict["proteins.proteinHalfLife.publication.pubMedId"] = row[
            "proteins.proteinHalfLife.publication.pubMedId"
        ]
        data_dict["organism.shortName"] = row["organism.shortName"]
        data_dict_list.append(data_dict)
    protein_half_life_dict["protein_half_life"] = data_dict_list
    # if protein_half_life_dict is empty set to None
    if protein_half_life_dict == {}:
        protein_half_life_dict["protein_half_life"] = None
    return protein_half_life_dict


# Retrieve the chromosomal location for a gene
def get_chromosomal_location(gene: str = "act1") -> dict:
    service = Service("https://yeastmine.yeastgenome.org/yeastmine/service")
    template = service.get_template("Gene_ChromosomeLocation")
    rows = template.rows(A={"op": "LOOKUP", "value": gene, "extra_value": ""})
    data_dict_list = []
    chromosomal_location_dict = {}
    for row in rows:
        data_dict = {}
        data_dict["primaryIdentifier"] = row["primaryIdentifier"]
        data_dict["secondaryIdentifier"] = row["secondaryIdentifier"]
        data_dict["symbol"] = row["symbol"]
        data_dict["name"] = row["name"]
        data_dict["organism.shortName"] = row["organism.shortName"]
        data_dict["chromosome.primaryIdentifier"] = row["chromosome.primaryIdentifier"]
        data_dict["chromosomeLocation.start"] = row["chromosomeLocation.start"]
        data_dict["chromosomeLocation.end"] = row["chromosomeLocation.end"]
        data_dict["chromosomeLocation.strand"] = row["chromosomeLocation.strand"]
        data_dict_list.append(data_dict)
    chromosomal_location_dict["chromosomal_location"] = data_dict_list
    # if chromosomal_location_dict is empty set to None
    if chromosomal_location_dict == {}:
        chromosomal_location_dict["chromosomal_location"] = None
    return chromosomal_location_dict


def get_feature_type(gene: str = "YIL080W") -> dict:
    service = Service("https://yeastmine.yeastgenome.org/yeastmine/service")
    # Get a new query on the class (table) you will be querying:
    query = service.new_query("Gene")
    # The view specifies the output columns
    query.add_view("featureType")
    # Uncomment and edit the line below (the default) to select a custom sort order:
    # query.add_sort_order("Gene.description", "ASC")
    query.add_constraint("organism.shortName", "=", "S. cerevisiae", code="F")
    query.add_constraint("Gene", "LOOKUP", gene, code="A")
    data_dict_list = []
    description_dict = {}
    for row in query.rows():
        data_dict = {}
        data_dict["featureType"] = row["featureType"]
        data_dict_list.append(data_dict)
    description_dict["featureType"] = data_dict_list
    # if description_dict is empty set to None
    if description_dict == {}:
        description_dict["featureType"] = None
    return description_dict


def get_all_feature_types() -> dict:
    service = Service("https://yeastmine.yeastgenome.org/yeastmine/service")
    # Get a new query on the class (table) you will be querying:
    query = service.new_query("Gene")
    # The view specifies the output columns
    query.add_view("featureType", "secondaryIdentifier")
    # Uncomment and edit the line below (the default) to select a custom sort order:
    # query.add_sort_order("Gene.featureType", "ASC")
    data_dict = {}
    for row in query.rows():
        data_dict[row["secondaryIdentifier"]] = row["featureType"]
    return data_dict


def get_pathways(gene: str = "fas1") -> dict:
    # Gene associated pathways
    # Retrieve all associated pathways for a specific gene. Gene(target) -> Gene Phenotypes.
    service = Service("https://yeastmine.yeastgenome.org/yeastmine/service")
    template = service.get_template("Gene_Pathways")
    rows = template.rows(A={"op": "LOOKUP", "value": gene, "extra_value": ""})
    data_dict_list = []
    pathways_dict = {}
    for row in rows:
        data_dict = {}
        data_dict["primaryIdentifier"] = row["primaryIdentifier"]
        data_dict["symbol"] = row["symbol"]
        data_dict["secondaryIdentifier"] = row["secondaryIdentifier"]
        data_dict["sgdAlias"] = row["sgdAlias"]
        data_dict["name"] = row["name"]
        data_dict["organism.shortName"] = row["organism.shortName"]
        data_dict["pathways.identifier"] = row["pathways.identifier"]
        data_dict["pathways.name"] = row["pathways.name"]
        data_dict_list.append(data_dict)
    pathways_dict["pathways"] = data_dict_list
    if pathways_dict == {}:
        pathways_dict["pathways"] = None
    return pathways_dict


def get_phenotype(gene: str = "act1") -> dict:
    service = Service("https://yeastmine.yeastgenome.org/yeastmine/service")
    template = service.get_template("Gene_Phenotype_New")
    rows = template.rows(
        A={"op": "LOOKUP", "value": gene, "extra_value": "S. cerevisiae"}
    )
    data_dict_list = []
    phenotype_dict = {}
    for row in rows:
        data_dict = {}
        data_dict["primaryIdentifier"] = row["primaryIdentifier"]
        data_dict["secondaryIdentifier"] = row["secondaryIdentifier"]
        data_dict["symbol"] = row["symbol"]
        data_dict["sgdAlias"] = row["sgdAlias"]
        data_dict["qualifier"] = row["qualifier"]
        data_dict["phenotypes.experimentType"] = row["phenotypes.experimentType"]
        data_dict["phenotypes.mutantType"] = row["phenotypes.mutantType"]
        data_dict["phenotypes.observable"] = row["phenotypes.observable"]
        data_dict["phenotypes.qualifier"] = row["phenotypes.qualifier"]
        data_dict["phenotypes.allele"] = row["phenotypes.allele"]
        data_dict["phenotypes.alleleDescription"] = row["phenotypes.alleleDescription"]
        data_dict["phenotypes.strainBackground"] = row["phenotypes.strainBackground"]
        data_dict["phenotypes.chemical"] = row["phenotypes.chemical"]
        data_dict["phenotypes.condition"] = row["phenotypes.condition"]
        data_dict["phenotypes.details"] = row["phenotypes.details"]
        data_dict["phenotypes.reporter"] = row["phenotypes.reporter"]
        data_dict["phenotypes.publications.pubMedId"] = row[
            "phenotypes.publications.pubMedId"
        ]
        data_dict["phenotypes.publications.citation"] = row[
            "phenotypes.publications.citation"
        ]
        data_dict_list.append(data_dict)
    phenotype_dict["phenotype"] = data_dict_list
    if phenotype_dict == {}:
        phenotype_dict["phenotype"] = None
    return phenotype_dict


# Was originally used to get the "symbol" gene names (like act1) for genes with standard names. Used for the Mechanisitc-Aware comparison. Not yet written to ym_attrs.
def get_gene_external_id(gene: str = "act1") -> dict:
    service = Service("https://yeastmine.yeastgenome.org/yeastmine/service")
    query = service.new_query("Gene")
    query.add_view(
        "primaryIdentifier",
        "secondaryIdentifier",
        "symbol",
        "name",
        "sgdAlias",
        "crossReferences.identifier",
        "crossReferences.source.name",
    )
    query.add_constraint("organism.shortName", "=", "S. cerevisiae", code="B")
    query.add_constraint("Gene", "LOOKUP", gene, code="A")
    data_dict_list = []
    external_id_dict = {}
    for row in query.rows():
        data_dict = {}
        data_dict["primaryIdentifier"] = row["primaryIdentifier"]
        data_dict["secondaryIdentifier"] = row["secondaryIdentifier"]
        data_dict["symbol"] = row["symbol"]
        data_dict["name"] = row["name"]
        data_dict["sgdAlias"] = row["sgdAlias"]
        data_dict["crossReferences.identifier"] = row["crossReferences.identifier"]
        data_dict["crossReferences.source.name"] = row["crossReferences.source.name"]
        data_dict_list.append(data_dict)

    external_id_dict["external_ids"] = data_dict_list
    if external_id_dict == {}:
        external_id_dict["external_ids"] = None
    return external_id_dict


def write_gene_id_translation_table() -> tuple[pd.DataFrame, list[str]]:
    # This function takes 2.5 hours to run. Could probably pull from somewhere else... but this is convenient in that it is specific to the gene_list.
    gene_list = get_gene_list()
    list_of_dicts = []
    for i, gene in enumerate(tqdm(gene_list)):
        errors = []
        try:
            symbol = get_gene_external_id(gene)["external_ids"][0]["symbol"]
        except IndexError as error:
            errors.append(f"gene: {error}")
        list_of_dicts.append({"secondaryIdentifier": gene, "symbol": symbol})
    df = pd.DataFrame(list_of_dicts)
    path = "data/preprocessed/gene_id_translation_table.csv"
    df.to_csv(path, index=False)
    print(f"Table writen to: {path}")
    return df, errors


def merge_dicts(dict_list: list[dict]) -> dict:
    result = {}
    for d in dict_list:
        result.update(d)
    return result


def get_ym_attrs_list(gene: str) -> list[dict]:
    regulators = delayed(get_regulators)(gene)
    physical_interaction = delayed(get_physical_interactions)(gene)
    gene_interaction = delayed(get_gene_interactions)(gene)
    phenotype = delayed(get_phenotype)(gene)
    protein_abundance = delayed(get_protein_abundance)(gene)
    median_protein_abundance = delayed(get_median_protein_abundance)(gene)
    protein_sequence = delayed(get_protein_sequence)(gene)
    gene_sequence = delayed(get_gene_sequence)(gene)
    go = delayed(get_go)(gene)
    protein_half_life = delayed(get_protein_half_life)(gene)
    chromosomal_location = delayed(get_chromosomal_location)(gene)
    feature_type = delayed(get_feature_type)(gene)
    pathways = delayed(get_pathways)(gene)

    ym_attrs_list = [
        regulators,
        physical_interaction,
        gene_interaction,
        phenotype,
        protein_abundance,
        median_protein_abundance,
        protein_sequence,
        gene_sequence,
        go,
        protein_half_life,
        chromosomal_location,
        feature_type,
        pathways,
    ]
    return ym_attrs_list


def create_node_ym_attrs(gene_list: list) -> dict:
    # Failures occur do to requests connections errors.
    file_names_list = []
    node_ym_attrs_dir = "data/preprocessed/node_ym_attrs"
    if not osp.exists(node_ym_attrs_dir):
        os.mkdir(node_ym_attrs_dir)
    else:
        # sort by date modified
        files = os.listdir(node_ym_attrs_dir)
        files.sort(key=lambda x: os.path.getmtime(osp.join(node_ym_attrs_dir, x)))
    file_names_list = [f.split(".")[0] for f in files]
    for gene in tqdm(gene_list):
        # Overwrite the last file in list, where previous run failed.
        if gene not in file_names_list[:-1]:
            print(gene)
            ym_attrs = compute(get_ym_attrs_list(gene))[0]
            ym_attrs_dict = {gene: merge_dicts(ym_attrs)}
            # write ym_attrs_list to json with files name indexed by ith gene
            with open(f"{node_ym_attrs_dir}/{gene}.json", "w") as f:
                json.dump(ym_attrs_dict, f, indent=4)
    return ym_attrs_dict


if __name__ == "__main__":
    # gene_list = get_gene_list()
    # create_node_ym_attrs(gene_list)
    ###
    # df, errors = write_gene_id_translation_table()
    # print(df)
    # print(errors)
    ###
    # get_gene_sequence()
    get_regulators()
    pass



# torchcell/models/esm2.py
# [[torchcell.models.esm2]]
# https://github.com/Mjvolk3/torchcell/tree/main/torchcell/models/esm2.py
# Test file: torchcell/models/test_esm2.py

import os
import os.path as osp

import torch
from transformers import AutoModelForMaskedLM, AutoTokenizer

from torchcell.models.llm import PeptideModel


class Esm2(PeptideModel):
    VALID_MODEL_NAMES = [
        "esm2_t48_15B_UR50D",
        "esm2_t36_3B_UR50D",
        "esm2_t33_650M_UR50D",
        "esm2_t30_150M_UR50D",
        "esm2_t12_35M_UR50D",
        "esm2_t6_8M_UR50D",
    ]

    def __init__(self, model_name: str):
        self.tokenizer = None
        self.model_name = osp.join("facebook", model_name)
        super().__init__(self.model_name)

    @staticmethod
    def _check_and_download_model(model_name: str):
        # Define the model name

        # Define the directory where you want the model to be saved
        script_dir = osp.dirname(osp.realpath(__file__))
        target_directory = osp.join(script_dir, "pretrained_LLM", "Esm2")

        # Create the target directory if it doesn't exist
        if not osp.exists(target_directory):
            os.makedirs(target_directory)

        model_directory = osp.join(target_directory, model_name)

        # Check if the model has already been downloaded
        if osp.exists(model_directory):
            print(f"{model_name} model already downloaded.")
        else:
            print(f"Downloading {model_name} model to {model_directory}...")
            # tokenizer
            AutoTokenizer.from_pretrained(model_name, cache_dir=target_directory)
            # model
            AutoModelForMaskedLM.from_pretrained(model_name, cache_dir=target_directory)
            print("Download finished.")

    @property
    def max_sequence_size(self):
        # TODO change.
        return 1026

    def load_model(self, model_name: str):
        # Check and download the model if necessary
        self._check_and_download_model(model_name)

        # Load the tokenizer and the model
        self.tokenizer = AutoTokenizer.from_pretrained(self.model_name)
        self.model = AutoModelForMaskedLM.from_pretrained(self.model_name)

    def embed(self, sequences: list[str], mean_embedding: bool = False) -> torch.Tensor:
        # Encode sequences and obtain attention masks directly
        encoding = self.tokenizer.batch_encode_plus(
            sequences,
            return_tensors="pt",
            padding=True,
            truncation=True,
            max_length=self.max_sequence_size,
        )
        tokens_ids = encoding["input_ids"]
        attention_mask = encoding["attention_mask"]

        # Check token lengths
        for token_id in tokens_ids:
            if len(token_id) > self.tokenizer.model_max_length:
                raise ValueError(
                    f"Token length {len(token_id)} exceeds the model's maximum length of {self.tokenizer.model_max_length}."
                )

        # Compute the embeddings
        torch_outs = self.model(
            tokens_ids, attention_mask=attention_mask, output_hidden_states=True
        )

        embeddings = torch_outs["hidden_states"][-1].detach()

        if mean_embedding:
            # Compute mean embeddings per sequence
            embeddings = (attention_mask.unsqueeze(-1) * embeddings).sum(
                dim=1
            ) / attention_mask.sum(dim=1).unsqueeze(-1)

        return embeddings


if __name__ == "__main__":
    model_names = [
        # "esm2_t48_15B_UR50D",
        # "esm2_t36_3B_UR50D",
        # "esm2_t33_650M_UR50D",
        # "esm2_t30_150M_UR50D",
        # "esm2_t12_35M_UR50D",
        "esm2_t6_8M_UR50D"
    ]
    for model_name in model_names:
        model = Esm2(model_name=model_name)
        sample_sequences = ["A" * 1024]
        embeddings = model.embed(sample_sequences, mean_embedding=True)
        print(embeddings.shape)



# torchcell/datasets/one_hot_gene.py
# [[torchcell.datasets.one_hot_gene]]
# https://github.com/Mjvolk3/torchcell/tree/main/torchcell/datasets/one_hot_gene.py
# Test file: torchcell/datasets/test_one_hot_gene.py

from collections.abc import Callable
import os
import torch
from torch_geometric.data import Data
from tqdm import tqdm

from torchcell.datasets.embedding import BaseEmbeddingDataset
from torchcell.sequence import GeneSet, ParsedGenome
from torchcell.sequence.genome.scerevisiae.s288c import SCerevisiaeGenome


class OneHotGeneDataset(BaseEmbeddingDataset):
    MODEL_TO_WINDOW = {"one_hot_gene": None}

    def __init__(
        self,
        root: str,
        genome: SCerevisiaeGenome,
        transform: Callable | None = None,
        pre_transform: Callable | None = None,
    ):
        # Create a mapping from gene to its index
        self.genome = genome
        self.gene_to_index = {
            gene: idx for idx, gene in enumerate(self.genome.gene_set)
        }

        self.model_name = "one_hot_gene"
        super().__init__(root, self.model_name, transform, pre_transform)
        self.genome = self.parse_genome(genome)
        del genome

        # HACK
        if self.model_name:
            if not os.path.exists(self.processed_paths[0]):
                self.transformer = self.initialize_transformer()
                self.process()
            self.data, self.slices = torch.load(self.processed_paths[0])

    # This is done to avoid pkl error when since genome uses sqlite
    @staticmethod
    def parse_genome(genome) -> ParsedGenome:
        # BUG we have to do this black magic because when you merge datasets with +
        # the genome is None
        if genome is None:
            return None
        else:
            data = {}
            data["gene_set"] = genome.gene_set
            return ParsedGenome(**data)

    def initialize_model(self):
        return None

    def one_hot_encode_gene(self, gene: str) -> torch.Tensor:
        """One-hot encode a gene based on its position in the gene set."""
        encoded = torch.zeros(len(self.gene_to_index))
        encoded[self.gene_to_index[gene]] = 1
        return encoded

    def process(self):
        # HACK
        if not self.model_name:
            return

        data_list = []

        for gene_id in tqdm(self.genome.gene_set):
            encoded_gene = self.one_hot_encode_gene(gene_id)

            # Create a Data object
            data = Data(id=gene_id, dna_windows={})
            data.embeddings = {self.model_name: encoded_gene.unsqueeze(0)}
            data_list.append(data)

        if self.pre_transform:
            data_list = [self.pre_transform(data) for data in data_list]

        torch.save(self.collate(data_list), self.processed_paths[0])


if __name__ == "__main__":
    import os
    import os.path as osp

    from dotenv import load_dotenv

    load_dotenv()
    DATA_ROOT = os.getenv("DATA_ROOT")

    genome = SCerevisiaeGenome(data_root=osp.join(DATA_ROOT, "data/sgd/genome"))

    genome = SCerevisiaeGenome()
    dataset = OneHotGeneDataset(
        root="data/scerevisiae/gene_one_hot_encoded", genome=genome
    )

    some_data = dataset[0]  # Should give you the first dataset item
    print(some_data)


if __name__ == "__main__":
    pass



# torchcell/datasets/ontology.py
# [[torchcell.datasets.ontology]]
# https://github.com/Mjvolk3/torchcell/tree/main/torchcell/datasets/ontology.py
# Test file: torchcell/datasets/test_ontology.py
import json

from owlready2 import (
    DataProperty,
    FunctionalProperty,
    ObjectProperty,
    Thing,
    get_ontology,
)

# Create a new ontology
# currently only have rdf
onto = get_ontology(
    "https://raw.githubusercontent.com/Mjvolk3/torchcell/main/torchcell.rdf"
)
from neo4j import GraphDatabase


# Define the top-level Experiment class
class Experiment(Thing):
    namespace = onto


# Define Genotype, Phenotype, and Environment as subclasses of Experiment
class Genotype(Experiment):
    namespace = onto


class Phenotype(Experiment):
    namespace = onto


class Environment(Experiment):
    namespace = onto


# Define properties for Genotype


class ReferenceGenome(Genotype):
    namespace = onto
    domain = [Genotype]
    range = [str]


class SysGeneName(Genotype):
    namespace = onto
    domain = [Genotype]
    range = [str]


class Perturbation(Genotype):
    namespace = onto
    domain = [Genotype]
    range = [str]


class GeneDeletion(Perturbation):
    namespace = onto
    domain = [Perturbation]
    range = [str]


class SysGeneNameFull(Genotype):
    namespace = onto
    domain = [Genotype]
    range = [str]


# Define properties for Phenotype
class Smf(Phenotype):
    namespace = onto
    domain = [Phenotype]
    range = [float]


# Define properties for Phenotype
class SmfStd(Phenotype):
    namespace = onto
    domain = [Phenotype]
    range = [float]


class Dmf(Phenotype):
    namespace = onto
    domain = [Phenotype]
    range = [float]


class DmfStd(Phenotype):
    namespace = onto
    domain = [Phenotype]
    range = [float]


class GeneticInteractionScore(Phenotype):
    namespace = onto
    domain = [Phenotype]
    range = [float]


class GeneticInteractionPValue(Phenotype):
    namespace = onto
    domain = [Phenotype]
    range = [float]


# Define properties for Environment
class Media(Environment):
    namespace = onto


class Chemical(Thing):
    namespace = onto


class YeastExtract(Chemical):
    namespace = onto


class Peptone(Chemical):
    namespace = onto


class Dextrose(Chemical):
    namespace = onto


class ComposedOf(ObjectProperty):
    namespace = onto
    domain = [Media]
    range = [Chemical]


class Ypd(Media):
    namespace = onto
    composed_of = ComposedOf()


class Temperature(Environment):
    namespace = onto
    domain = [Environment]
    range = [int]


def create_unique_constraint_if_not_exists(driver):
    with driver.session() as session:
        # Get existing constraints
        constraints = session.run("SHOW CONSTRAINTS").data()

        # Check if the specific constraint exists
        if not any(
            "n10s_unique_uri" in constraint.get("name", "")
            for constraint in constraints
        ):
            # Create the constraint if it does not exist
            session.run(
                "CREATE CONSTRAINT n10s_unique_uri FOR (r:Resource) REQUIRE r.uri IS UNIQUE"
            )


def owl_import_ex():
    # Connection details
    uri = "neo4j://localhost:7687"  # Adjust as needed
    username = "neo4j"
    password = "torchcell"

    # Connect to Neo4j
    driver = GraphDatabase.driver(uri, auth=(username, password))

    # Create the unique constraint if it does not exist
    create_unique_constraint_if_not_exists(driver)

    with driver.session() as session:
        # Execute the n10s graph configuration initialization
        session.run("CALL n10s.graphconfig.init();")

        # Execute the ontology import
        session.run(
            "CALL n10s.onto.import.fetch("
            "'https://raw.githubusercontent.com/Mjvolk3/torchcell/main/torchcell.rdf', "
            "'RDF/XML');"
        )
        print("Ontology import completed successfully.")

    # Close the driver connection
    driver.close()


def main():
    # Create instances of YeastExtract, Peptone, and Dextrose
    yeast_extract = YeastExtract()
    peptone = Peptone()
    dextrose = Dextrose()

    # Create an instance of Ypd and link its components
    ypd = Ypd()
    ypd.composed_of = [yeast_extract, peptone, dextrose]

    onto.save(file="torchcell.rdf", format="rdfxml")


if __name__ == "__main__":
    main()
    owl_import_ex()



# torchcell/sequence/__init__.py
from .data import (
    DnaSelectionResult,
    DnaWindowResult,
    Gene,
    GeneSet,
    Genome,
    ParsedGenome,
    calculate_window_bounds,
    calculate_window_bounds_symmetric,
    compute_codon_frequency,
    get_chr_from_description,
    mismatch_positions,
    roman_to_int,
)

data_classes = [
    "DnaSelectionResult",
    "DnaWindowResult",
    "Gene",
    "GeneSet",
    "Genome",
    "ParsedGenome",
]

helper_functions = [
    "calculate_window_bounds",
    "calculate_window_bounds_symmetric",
    "get_chr_from_description",
    "mismatch_positions",
    "roman_to_int",
    "compute_codon_frequency",
]

__all__ = data_classes + helper_functions



# torchcell/datasets/scerevisiae/baryshnikovna2010.py
# [[torchcell.datasets.scerevisiae.baryshnikovna2010]]
# https://github.com/Mjvolk3/torchcell/tree/main/torchcell/datasets/scerevisiae/baryshnikovna2010.py
# Test file: torchcell/datasets/scerevisiae/test_baryshnikovna2010.py
import os
from typing import Callable, Optional

import pandas as pd
import torch
from torch_geometric.data import Data, InMemoryDataset, download_url


os.makedirs("data/scerevisiae/baryshnikovna2010", exist_ok=True)


class Baryshnikovna2010Dataset(InMemoryDataset):
    url = "https://static-content.springer.com/esm/art%3A10.1038%2Fnmeth.1534/MediaObjects/41592_2010_BFnmeth1534_MOESM168_ESM.xls"

    def __init__(
        self,
        root: str = "data/scerevisiae/baryshnikovna2010",
        transform: Optional[Callable] = None,
        pre_transform: Optional[Callable] = None,
    ):
        super().__init__(root, transform, pre_transform)
        self.data, self.slices = torch.load(self.processed_paths[0])

    @property
    def raw_file_names(self) -> list[str]:
        return ["data.xls"]

    @property
    def processed_file_names(self) -> str:
        return "data.pt"

    def download(self):
        path = download_url(self.url, self.raw_dir)
        os.rename(path, os.path.join(self.raw_dir, "data.xls"))

    def process(self):
        xls_path = os.path.join(self.raw_dir, "data.xls")
        df = pd.read_excel(xls_path, names=["id", "fitness", "std"])

        data_list = []
        for index, row in df.iterrows():
            id = row["id"].split("_")[0]
            id_full = row["id"]
            genotype = {
                "id": id,
                "interventions": "deletion",
                "id_full": id_full,
            }
            observation = {
                "smf_fitness": torch.tensor([row["fitness"]], dtype=torch.float),
                "smf_std": torch.tensor([row["std"]], dtype=torch.float),
            }
            environment = {
                "media": "YPD",
                "temperature": "30C",
            }
            phenotype = {"observation": observation, "environment": environment}

            data = Data()
            data.genotype = genotype
            data.phenotype = phenotype
            data_list.append(data)

        if self.pre_transform:
            data_list = [self.pre_transform(data) for data in data_list]

        # Save the processed data
        torch.save(self.collate(data_list), self.processed_paths[0])


if __name__ == "__main__":
    dataset = Baryshnikovna2010Dataset(root="data/scerevisiae/baryshnikovna2010")
    print(40 * "=")
    print("Lets take a quick look at the dataset...")
    print(f">>> dataset\n{dataset}")
    print(f">>> dataset[0]\n{dataset[0]}")
    print(f">>> dataset[0].genotype\n{dataset[0].genotype}")
    print(f">>> dataset[0].phenotype\n{dataset[0].phenotype}")
    print(f">>> dataset[0].environment\n{dataset[0].environment}")
    print(40 * "=")



# torchcell/knowledge_graphs/create_scerevisiae_kg_small
# [[torchcell.knowledge_graphs.create_scerevisiae_kg_small]]
# https://github.com/Mjvolk3/torchcell/tree/main/torchcell/knowledge_graphs/create_scerevisiae_kg_small.py
# Test file: tests/torchcell/knowledge_graphs/test_create_scerevisiae_kg_small.py

from biocypher import BioCypher
from torchcell.adapters import (
    SmfCostanzo2016Adapter,
    DmfCostanzo2016Adapter,
    SmfKuzmin2018Adapter,
    DmfKuzmin2018Adapter,
    TmfKuzmin2018Adapter,
)
from torchcell.datasets.scerevisiae import (
    SmfCostanzo2016Dataset,
    DmfCostanzo2016Dataset,
    SmfKuzmin2018Dataset,
    DmfKuzmin2018Dataset,
    TmfKuzmin2018Dataset,
)
import logging
from dotenv import load_dotenv
import os
import os.path as osp
from datetime import datetime
import multiprocessing as mp
import sys
import math

logging.basicConfig(level=logging.INFO)
log = logging.getLogger(__name__)

# WARNING do not print in this file! This file is used to generate a path to a bash script and printing to stdout will break the bash script path


def get_num_workers():
    """Get the number of CPUs allocated by SLURM."""
    # Try to get number of CPUs allocated by SLURM
    cpus_per_task = os.getenv("SLURM_CPUS_PER_TASK")
    if cpus_per_task is not None:
        return int(cpus_per_task)
    # Fallback: Use multiprocessing to get the total number of CPUs
    return mp.cpu_count()


def main() -> str:
    # Configure logging
    logging.basicConfig(level=logging.INFO, filename="biocypher_warnings.log")
    logging.captureWarnings(True)
    load_dotenv()
    DATA_ROOT = os.getenv("DATA_ROOT")
    BIOCYPHER_CONFIG_PATH = os.getenv("BIOCYPHER_CONFIG_PATH")
    SCHEMA_CONFIG_PATH = os.getenv("SCHEMA_CONFIG_PATH")
    BIOCYPHER_OUT_PATH = os.getenv("BIOCYPHER_OUT_PATH")

    # Use this function to get the number of workers
    num_workers = get_num_workers()
    time = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
    log.info(f"Number of workers: {num_workers}")
    bc = BioCypher(
        output_directory=osp.join(DATA_ROOT, BIOCYPHER_OUT_PATH, time),
        biocypher_config_path=BIOCYPHER_CONFIG_PATH,
        schema_config_path=SCHEMA_CONFIG_PATH,
    )
    # Partition workers
    num_workers = get_num_workers()
    io_workers = math.ceil(0.2 * num_workers)
    compute_workers = num_workers - io_workers
    chunk_size = int(1e5)
    loader_batch_size = int(1e3)

    # Ordered adapters from smallest to largest
    adapters = [
        SmfCostanzo2016Adapter(
            dataset=SmfCostanzo2016Dataset(
                root=osp.join(DATA_ROOT, "data/torchcell/smf_costanzo2016")
            ),
            compute_workers=compute_workers,
            io_workers=io_workers,
            chunk_size=chunk_size,
            loader_batch_size=loader_batch_size,
        ),
        DmfCostanzo2016Adapter(
            dataset=DmfCostanzo2016Dataset(
                root=osp.join(DATA_ROOT, "data/torchcell/dmf_costanzo2016_1e5"),
                subset_n=int(1e5),
            ),
            compute_workers=compute_workers,
            io_workers=io_workers,
            chunk_size=chunk_size,
            loader_batch_size=loader_batch_size,
        ),
        SmfKuzmin2018Adapter(
            dataset=SmfKuzmin2018Dataset(
                root=osp.join(DATA_ROOT, "data/torchcell/smf_kuzmin2018")
            ),
            compute_workers=compute_workers,
            io_workers=io_workers,
            chunk_size=chunk_size,
            loader_batch_size=loader_batch_size
        ),
        DmfKuzmin2018Adapter(
            dataset=DmfKuzmin2018Dataset(
                root=osp.join(DATA_ROOT, "data/torchcell/dmf_kuzmin2018")
            ),
            compute_workers=compute_workers,
            io_workers=io_workers,
            chunk_size=chunk_size,
            loader_batch_size=loader_batch_size
        ),
        TmfKuzmin2018Adapter(
            dataset=TmfKuzmin2018Dataset(
                root=osp.join(DATA_ROOT, "data/torchcell/tmf_kuzmin2018")
            ),
            compute_workers=compute_workers,
            io_workers=io_workers,
            chunk_size=chunk_size,
            loader_batch_size=loader_batch_size
        ),
    ]

    for adapter in adapters:
        bc.write_nodes(adapter.get_nodes())
        bc.write_edges(adapter.get_edges())

    log.info("Finished iterating nodes and edges")
    # Write admin import statement and schema information (for biochatter)
    bc.write_import_call()
    bc.write_schema_info(as_node=True)

    # bc.summary()
    # Returns bash script path

    relative_bash_script_path = osp.join(
        "biocypher-out", time, "neo4j-admin-import-call.sh"
    )
    return relative_bash_script_path


if __name__ == "__main__":
    print(main())



# torchcell/cell/cell.py
import os
import shutil
import zipfile
from abc import ABC, abstractmethod
from collections.abc import Callable
from os import environ
from typing import List, Optional, Tuple, Union

import pandas as pd
import torch
from attrs import define
from torch_geometric.data import Data, InMemoryDataset, download_url, extract_zip

from torchcell.data_prior.sequence import SCerevisiaeGenome
from torchcell.datasets.scerevisiae import (
    DmfCostanzo2016Dataset,
    SmfCostanzo2016Dataset,
)
from torchcell.sequence import Genome


class DiMultiGraph:
    pass


class Ontology:
    """
    Represents the biological ontology guiding the data join.
    """

    def __init__(self):
        # TODO: Initialize ontology
        pass

    def join(self, other: "Ontology") -> "Ontology":
        # TODO: Implement ontology join logic
        pass


class CellDataset(InMemoryDataset):
    """
    Represents a dataset for cellular data.
    """

    def __init__(
        self,
        root: str,
        transform: Callable | None = None,
        pre_transform: Callable | None = None,
        pre_filter: Callable | None = None,
    ):
        super().__init__(root, transform, pre_transform, pre_filter)
        self.genome: Genome
        self.dimultigraph: DiMultiGraph | None = None
        self.experiment_datasets: list[InMemoryDataset] | None = None
        self.ontology: Ontology | None = None

    @property
    def processed_file_names(self) -> list[str]:
        return ["cell.pt"]

    def process():
        pass

    def __and__(self, other: "CellDataset") -> "IntersectionDataset":
        return IntersectionDataset(self, other)


class IntersectionDataset(CellDataset):
    """
    Represents a dataset formed by joining two CellDatasets.
    """

    def __init__(self, ds1: CellDataset, ds2: CellDataset):
        # TODO: Logic to initialize the IntersectionDataset using ds1 and ds2
        pass

    def _merge_dataset_indices(self):
        # TODO: Logic to merge indices from the two datasets
        pass

    def __getitem__(self, idx: int) -> Data:
        # TODO: Logic to get item given an index
        pass


# Convenience functions as described in the notes:


def visualize_ontology(ontology: Ontology):
    """
    Visualizes the given ontology.
    """
    # TODO: Implement visualization
    pass


def compare_ontologies(ont1: Ontology, ont2: Ontology):
    """
    Compares two ontologies and returns overlapping and conflicting parts.
    """
    # TODO: Implement comparison
    pass


def recall_dropped_data(dataset: IntersectionDataset):
    """
    Recalls any data that was dropped during the join.
    """
    # TODO: Implement recall
    pass


def recall_transformed_data(dataset: IntersectionDataset):
    """
    Recalls any data that was transformed during the join.
    """
    # TODO: Implement recall
    pass


# The main function and entry point:


def main():
    print("cell loading")
    genome = SCerevisiaeGenome()
    cell = CellDataset(root="data/scerevisiae")
    cell.genome = genome
    cell.experiment_datasets = [DmfCostanzo2016Dataset()]


if __name__ == "__main__":
    main()



# torchcell/datasets/fungal_up_down_transformer.py
# [[torchcell.datasets.fungal_up_down_transformer]]
# https://github.com/Mjvolk3/torchcell/tree/main/torchcell/datasets/fungal_up_down_transformer.py
# Test file: torchcell/datasets/test_fungal_up_down_transformer.py

import os
from collections.abc import Callable
from typing import Optional

import torch
from torch_geometric.data import Data
from tqdm import tqdm

from torchcell.datasets.embedding import BaseEmbeddingDataset
from torchcell.models.fungal_up_down_transformer import (  # adjusted import
    FungalUpDownTransformer,
)
from torchcell.sequence import GeneSet, ParsedGenome, compute_codon_frequency
from torchcell.sequence.genome.scerevisiae.s288c import SCerevisiaeGenome


class CodonFrequencyDataset(BaseEmbeddingDataset):
    # Could add frequency for other parts of sequence
    # but this doesn't make much sense
    MODEL_TO_WINDOW = {"cds_codon_frequency": None}

    def __init__(
        self,
        root: str,
        genome: SCerevisiaeGenome,
        transform: Callable | None = None,
        pre_transform: Callable | None = None,
    ):
        self.genome = genome

        self.model_name = "cds_codon_frequency"
        super().__init__(root, self.model_name, transform, pre_transform)
        self.genome = self.parse_genome(genome)
        del genome

    # This is done to avoid pkl error when since genome uses sqlite
    @staticmethod
    def parse_genome(genome) -> ParsedGenome:
        # BUG we have to do this black magic because when you merge datasets with +
        # the genome is None
        if genome is None:
            return None
        else:
            data = {}
            data["gene_set"] = genome.gene_set
            return ParsedGenome(**data)

    def initialize_model(self):
        return None

    def process(self):
        data_list = []

        for gene_id in tqdm(self.genome.gene_set):
            sequence = str(self.genome[gene_id].cds.seq)

            # Check if the sequence is valid for codon frequency computation
            try:
                codon_frequency = compute_codon_frequency(sequence)
            except ValueError:
                continue

            # Create a Data object
            data = Data(id=gene_id, dna_windows={})
            data.embeddings = {
                self.model_name: torch.tensor(codon_frequency.values()).unsqueeze(0)
            }
            data_list.append(data)

        if self.pre_transform:
            data_list = [self.pre_transform(data) for data in data_list]

        torch.save(self.collate(data_list), self.processed_paths[0])


if __name__ == "__main__":
    genome = SCerevisiaeGenome()
    dataset = CodonFrequencyDataset(
        root="data/scerevisiae/codon_frequency", genome=genome
    )

    some_data = dataset[0]  # Should give you the first dataset item
    print(some_data)



import json
import os
from collections import defaultdict
from typing import Any

import matplotlib.pyplot as plt
import networkx as nx
import pydot
from adjustText import adjust_text
from networkx.drawing.nx_agraph import graphviz_layout
from networkx.drawing.nx_pydot import graphviz_layout


def analyze_structure(data, parent_key="") -> dict[str, Any]:
    type_structure = defaultdict(set)

    if isinstance(data, dict):
        for key, value in data.items():
            full_key = f"{parent_key}.{key}" if parent_key else key
            type_structure.update(analyze_structure(value, full_key))
    elif isinstance(data, list):
        for i, value in enumerate(data):
            if isinstance(value, (list, dict)):
                type_structure.update(analyze_structure(value, f"{parent_key}[{i}]"))
            else:
                type_structure[f"{parent_key}[{i}]"].add(type(value).__name__)
    else:
        type_structure[parent_key].add(type(data).__name__)

    return type_structure


def build_graph(data, graph=None):
    if graph is None:
        graph = nx.DiGraph()

    if isinstance(data, dict):
        for key, value in data.items():
            key_parts = key.split(".")
            for i in range(len(key_parts)):
                # Add node to graph if it doesn't exist already
                if key_parts[i] not in graph.nodes():
                    graph.add_node(key_parts[i], label=key_parts[i])

                # Add edges between consecutive nodes in the path
                if i > 0:  # Skip root node because it has no parent
                    graph.add_edge(key_parts[i - 1], key_parts[i])

                # Add edges between sibling nodes
                if i < len(key_parts) - 1:  # Skip leaf nodes
                    sibling_keys = [
                        sibling_key
                        for sibling_key in data.keys()
                        if sibling_key.startswith(".".join(key_parts[: i + 1]) + ".")
                        and sibling_key != key  # Exclude the current key
                    ]
                    for sibling_key in sibling_keys:
                        sibling_node = sibling_key.split(".")[i + 1]
                        if sibling_node not in graph.nodes():
                            graph.add_node(sibling_node, label=sibling_node)
                        graph.add_edge(key_parts[i], sibling_node)

            # The last node in the path (representing the type) is a leaf node
            leaf_node = str(value)  # Convert list to string using str function
            if leaf_node not in graph.nodes():
                graph.add_node(leaf_node, label=leaf_node)
            graph.add_edge(key_parts[-1], leaf_node)

    return graph


def calculate_depths(graph):
    if not graph.nodes:  # if the graph is empty
        return {}

    depths = {node: 0 for node in nx.nodes(graph)}  # Initialize depths to 0

    # Find the root of the graph (node with minimal in-degree)
    root = min(graph.nodes, key=graph.in_degree)

    # Perform breadth-first search starting from root
    queue = [(root, 0)]
    while queue:
        node, depth = queue.pop(0)
        depths[node] = depth
        queue.extend((child, depth + 1) for child in graph.successors(node))

    return depths


def draw_and_save_graph(graph, output_path, layout="spring"):
    plt.figure(figsize=(40, 30))  # Adjust as necessary

    if layout == "spring":
        pos = nx.spring_layout(graph, seed=42, center=[0.5, 0.5], k=1.0)  # k=1.5
    elif layout == "circular":
        pos = nx.nx_agraph.graphviz_layout(graph, prog="twopi")
    elif layout == "tree":
        pos = nx.nx_agraph.graphviz_layout(graph, prog="dot")

    depths = calculate_depths(graph)

    max_depth = max(depths.values()) if depths else 1  # Avoid division by zero
    node_list = list(graph.nodes())
    node_colors = [plt.cm.Greys(1.0 - depths[node] / max_depth) for node in node_list]

    # Overwrite leaf node colors with red
    leaf_nodes = [node for node, degree in graph.out_degree() if degree == 0]
    for node in leaf_nodes:
        node_colors[node_list.index(node)] = "red"

    nx.draw_networkx_nodes(
        graph, pos, node_color=node_colors, node_size=200
    )  # Adjust as necessary
    nx.draw_networkx_edges(graph, pos, edge_color="gray", alpha=0.6, arrows=True)

    node_labels = nx.get_node_attributes(graph, "label")
    texts = []
    for node, (x, y) in pos.items():
        texts.append(plt.text(x, y, node_labels[node], fontsize=16))

    if texts:  # Check if the list is not empty
        adjust_text(texts)

    plt.savefig(f"{output_path}.png", format="png", dpi=300)
    plt.savefig(f"{output_path}.pdf", format="pdf", dpi=300)
    plt.close()


def get_gene_name(file_path):
    return file_path.split("/")[-1].split(".")[0]


def process_src_to_image(path):
    split_path = path.split("/")
    index_of_src = split_path.index("src")
    post_src_path = split_path[index_of_src:]  # taking all parts after 'src'
    processed_path = "/".join(post_src_path)
    return processed_path


def analyze_json(file_path, create_images=False, layout="spring"):
    with open(file_path) as f:
        data = json.load(f)

    gene_name = get_gene_name(file_path)
    base_dir = os.path.dirname(os.path.abspath(__file__))
    image_path = process_src_to_image(base_dir)

    for key in data.keys():
        type_structure = analyze_structure(data[key])
        reduced_type_structure = {
            f"{key}.{k}": list(types) for k, types in type_structure.items()
        }

        # Dynamically construct the subdirectories for JSON files and images
        subdirs = key.split("/")  # Assuming 'key' is a path like "validation/locus"

        json_dir = os.path.join("data", "sgd", "validation", *subdirs)
        os.makedirs(json_dir, exist_ok=True)

        with open(os.path.join(json_dir, f"{gene_name}.json"), "w") as f:
            json.dump(reduced_type_structure, f, indent=4)

        if create_images:
            graph = build_graph(reduced_type_structure)  # Remove 'key' here

            image_dir = os.path.join("notes", "assets", "images", image_path, *subdirs)
            os.makedirs(image_dir, exist_ok=True)

            output_path = os.path.join(image_dir, f"{gene_name}_{key}_type_graph")
            draw_and_save_graph(graph, output_path, layout)


def main():
    import os

    file_path = "data/sgd/genes/YPR201W.json"
    analyze_json(file_path, create_images=True, layout="tree")


if __name__ == "__main__":
    main()



# torchcell/datasets/dcell.py
# [[torchcell.datasets.dcell]]
# https://github.com/Mjvolk3/torchcell/tree/main/torchcell/datasets/dcell.py
# Test file: torchcell/datasets/test_dcell.py
import json
import logging
import os
import os.path as osp
import pickle
from collections.abc import Callable

import lmdb
import networkx as nx
import numpy as np
import torch
from pydantic import field_validator
from torch_geometric.data import Data, InMemoryDataset
from torch_geometric.utils import (
    add_self_loops,
    from_networkx,
)
from tqdm import tqdm

from torchcell.dataset import Dataset
from torchcell.datamodels import ModelStrictArbitrary
from torchcell.datasets.embedding import BaseEmbeddingDataset
from torchcell.datasets.scerevisiae import (
    DmfCostanzo2016Dataset,
    SmfCostanzo2016Dataset,
)
from torchcell.graph import (
    filter_by_contained_genes,
    filter_by_date,
    filter_go_IGI,
    filter_redundant_terms,
)
from torchcell.models import DCell, DCellLinear, dcell, dcell_from_networkx
from torchcell.sequence import GeneSet, Genome
from torchcell.sequence.genome.scerevisiae.s288c import SCerevisiaeGenome

log = logging.getLogger(__name__)


class ParsedGenome(ModelStrictArbitrary):
    gene_set: GeneSet

    @field_validator("gene_set")
    def validate_gene_set(cls, value):
        if not isinstance(value, GeneSet):
            raise ValueError(f"gene_set must be a GeneSet, got {type(value).__name__}")
        return value


class DCellDataset(Dataset):
    """
    Represents a dataset for cellular data.
    """

    # TODO type change experiments
    def __init__(
        self,
        root: str = "data/scerevisiae/cell",
        # genome: Genome = None,
        genome: Genome = None,
        graph: nx.Graph = None,
        embeddings: BaseEmbeddingDataset | None = None,
        experiments: list[InMemoryDataset] | InMemoryDataset = None,
        zero_pert: bool = False,
        transform: Callable | None = None,
        pre_transform: Callable | None = None,
        pre_filter: Callable | None = None,
    ):
        self._gene_set = None
        self.graph = graph
        self.embeddings = embeddings
        self.experiments = experiments
        self.experiment_datasets = None

        # HACK zero out embedding of pert hack
        self.zero_pert = zero_pert

        # TODO consider moving to Dataset
        self.preprocess_dir = osp.join(root, "preprocess")

        # This is here because we can't run process without getting gene_set
        super().__init__(root, transform, pre_transform, pre_filter)
        # HACK start
        # Extract data from genome object, then remove for pickling with data loader
        self.genome = self.parse_genome(genome)
        del genome
        # HACK end
        # Create the seq graph
        graphs = []
        if self.embeddings:
            G_embedding = self.create_embedding_graph(self.genome, self.embeddings)
            graphs.append(G_embedding)
        self.cell_graph = self.to_cell_data(graphs)
        # HACK to try and get ride of no edge index issue.
        # Self loops fixes the batching problem for no edges.
        self.cell_graph.edge_index = add_self_loops(self.cell_graph.edge_index)[0]
        # LMDB env
        self.env = None

    def to_cell_data(self, graphs: list[nx.Graph]) -> Data:
        G = self.safe_compose(graphs)
        # drop nodes that don't belong to genome.gene_set
        data = from_networkx(G)
        data.ids = list(G.nodes())
        return data

    @staticmethod
    def parse_genome(genome) -> ParsedGenome:
        data = {}
        data["gene_set"] = genome.gene_set
        return ParsedGenome(**data)

    @property
    def raw_file_names(self) -> list[str]:
        # TODO consider return the processed of the experiments, etc.
        # This might cause an issue because there is expected behavior for raw,# and this is not it.
        return None  # Specify raw files if needed

    @property
    def processed_file_names(self) -> list[str]:
        return "data.lmdb"

    # HACK comment out for now
    # @property
    # def wt(self):
    #     # Need to be able to combine WTs into one WT
    #     # wts = [experiment.wt for experiment in self.experiments]
    #     # TODO aggregate WTS. For now just return the first one.
    #     wt = self.experiments.wt
    #     subset_data = self._subset_graph(wt)
    #     data = self._add_label(subset_data, wt)
    #     return data

    @property
    def wt(self):
        return None

    @staticmethod
    def safe_compose(graphs):
        if any(isinstance(G, nx.DiGraph) for G in graphs):
            # Convert all graphs to DiGraph if at least one is directed
            graphs = [
                G if isinstance(G, nx.DiGraph) else G.to_directed() for G in graphs
            ]

        # Start with an empty graph of the appropriate type
        composed_graph = (
            nx.DiGraph() if isinstance(graphs[0], nx.DiGraph) else nx.Graph()
        )

        for G in graphs:
            # Check for overlapping node data
            for node, data in G.nodes(data=True):
                if node in composed_graph:
                    for key, value in data.items():
                        if key in composed_graph.nodes[node]:
                            if isinstance(value, np.ndarray):
                                if not np.array_equal(
                                    value, composed_graph.nodes[node][key]
                                ):
                                    raise ValueError(
                                        f"Overlapping node data found for node {node}: {key}"
                                    )
                            elif composed_graph.nodes[node][key] != value:
                                raise ValueError(
                                    f"Overlapping node data found for node {node}: {key}"
                                )

            # Check for overlapping edge data
            for node1, node2, data in G.edges(data=True):
                if composed_graph.has_edge(node1, node2):
                    for key, value in data.items():
                        if key in composed_graph.edges[node1, node2]:
                            if isinstance(value, np.ndarray):
                                if not np.array_equal(
                                    value, composed_graph.edges[node1, node2][key]
                                ):
                                    raise ValueError(
                                        f"Overlapping edge data found for edge {(node1, node2)}: {key}"
                                    )
                            elif composed_graph.edges[node1, node2][key] != value:
                                raise ValueError(
                                    f"Overlapping edge data found for edge {(node1, node2)}: {key}"
                                )

            composed_graph = nx.compose(composed_graph, G)
        # After all graphs are composed unify nodes attrs into x
        # probably need to unify edge attrs too
        for node, data in composed_graph.nodes(data=True):
            attributes_list = []
            for key, value in data.items():
                if isinstance(value, np.ndarray):
                    attributes_list.append(value)
                # You can handle other types as needed

            # For simplicity, assuming all attributes are numpy arrays
            concatenated_attributes = np.concatenate(attributes_list)

            # Set the concatenated attributes to 'x' and remove other attributes
            composed_graph.nodes[node]["x"] = concatenated_attributes
            keys_to_remove = [key for key in data.keys() if key != "x"]
            for key in keys_to_remove:
                del composed_graph.nodes[node][key]

        return composed_graph

    @staticmethod
    def create_embedding_graph(genome, embeddings: BaseEmbeddingDataset) -> nx.Graph:
        """
        Create a NetworkX graph from embeddings.
        """
        # Create an empty NetworkX graph
        G = nx.Graph()

        # Extract and concatenate embeddings for all items in embeddings
        for item in embeddings:
            keys = item["embeddings"].keys()
            if item.id in genome.gene_set:
                item_embeddings = [item["embeddings"][k].squeeze(0) for k in keys]
                concatenated_embedding = torch.cat(item_embeddings)

                # Add nodes to the graph with embeddings as node attributes
                G.add_node(item.id, embedding=concatenated_embedding.numpy())

        return G

    def process(self):
        combined_data = []
        self.gene_set = self.compute_gene_set()

        # Precompute gene_set for faster lookup
        gene_set = self.gene_set

        # # Use list comprehension and any() for fast filtering
        combined_data = [
            item
            for item in tqdm(self.experiments)
            if all(i["id"] in gene_set for i in item.genotype)
        ]

        log.info("creating lmdb database")
        # Initialize LMDB environment
        env = lmdb.open(osp.join(self.processed_dir, "data.lmdb"), map_size=int(1e12))

        with env.begin(write=True) as txn:
            for idx, item in tqdm(enumerate(combined_data)):
                data = Data()
                data.genotype = item["genotype"]
                data.phenotype = item["phenotype"]

                # Serialize the data object using pickle
                serialized_data = pickle.dumps(data)

                # Save the serialized data in the LMDB environment
                txn.put(f"{idx}".encode(), serialized_data)

    @property
    def gene_set(self):
        try:
            if osp.exists(osp.join(self.preprocess_dir, "gene_set.json")):
                with open(osp.join(self.preprocess_dir, "gene_set.json")) as f:
                    self._gene_set = set(json.load(f))
            elif self._gene_set is None:
                raise ValueError(
                    "gene_set not written during process. "
                    "Please call compute_gene_set in process."
                )
            return GeneSet(self._gene_set)
        except json.JSONDecodeError:
            raise ValueError("Invalid or empty JSON file found.")

    @gene_set.setter
    def gene_set(self, value):
        if not value:
            raise ValueError("Cannot set an empty or None value for gene_set")
        if not osp.exists(self.preprocess_dir):
            os.makedirs(self.preprocess_dir)
        with open(osp.join(self.preprocess_dir, "gene_set.json"), "w") as f:
            json.dump(list(sorted(value)), f, indent=0)
        self._gene_set = value

    def compute_gene_set(self):
        if not self._gene_set:
            if isinstance(self.experiments, Dataset):
                experiment_gene_set = self.experiments.gene_set
            else:
                # TODO: handle other data types for experiments, if necessary
                raise NotImplementedError(
                    "Expected 'experiments' to be of type InMemoryDataset"
                )
            # Not sure we should take the intersection here...
            # Could use gene_set from genome instead, since this is base
            # In case of gene addition would need to update the gene_set
            # then cell_dataset should be max possible.
            cell_gene_set = set(self.genome.gene_set).intersection(experiment_gene_set)
        return cell_gene_set

    def _subset_graph(self, data: Data) -> Data:
        """
        Subset the reference graph based on the genes in data.genotype.
        """
        # Nodes to remove based on the genes in data.genotype
        nodes_to_remove = torch.tensor(
            [
                self.cell_graph.ids.index(gene["id"])
                for gene in data.genotype
                if gene["id"] in self.cell_graph.ids
            ],
            dtype=torch.long,
        )

        perturbed_nodes = nodes_to_remove.clone().detach()

        # Compute the nodes to keep
        all_nodes = torch.arange(self.cell_graph.num_nodes, dtype=torch.long)
        nodes_to_keep = torch.tensor(
            [node for node in all_nodes if node not in perturbed_nodes],
            dtype=torch.long,
        )

        # Get the induced subgraph using the nodes to keep
        subset_graph = self.cell_graph.subgraph(nodes_to_keep)
        subset_remove_graph = self.cell_graph.subgraph(nodes_to_remove)
        subset_graph.ids_pert = subset_remove_graph.ids
        # TODO remove all below.
        subset_graph.x_pert = subset_remove_graph.x
        subset_graph.x_pert_idx = perturbed_nodes
        # Dcell data
        G_dcell = dcell.delete_genes(
            go_graph=self.graph.G_go, deletion_gene_set=GeneSet(subset_graph.ids_pert)
        )
        data = dcell_from_networkx(G_dcell)
        return data

    def _add_label(self, data: Data, original_data: Data) -> Data:
        """
        Adds the dmf_fitness label to the data object if it exists in the original data's phenotype["observation"].

        Args:
            data (Data): The Data object to which the label should be added.
            original_data (Data): The original Data object from which the label should be extracted.

        Returns:
            Data: The modified Data object with the added label.
        """
        if "dmf" in original_data.phenotype["observation"]:
            # TODO change dmf in costanzo to be fitness - need to standardize
            data.fitness = original_data.phenotype["observation"]["dmf"]
        if "fitness" in original_data.phenotype["observation"]:
            # TODO change dmf in costanzo to be fitness - need to standardize
            data.fitness = original_data.phenotype["observation"]["fitness"]
        if "genetic_interaction_score" in original_data.phenotype["observation"]:
            data.genetic_interaction_score = original_data.phenotype["observation"][
                "genetic_interaction_score"
            ]
        return data

    # def get(self, idx: int) -> Data:
    #     env = lmdb.open(self.processed_paths[0], readonly=True, lock=False)
    #     with env.begin() as txn:
    #         serialized_data = txn.get(f"{idx}".encode())
    #         if serialized_data is None:
    #             return None
    #         data = pickle.loads(serialized_data)
    #         if self.transform:
    #             data = self.transform(data)

    #         # Get the subset data using the separate method
    #         subset_data = self._subset_graph(data)

    #         # Add the dmf_fitness label to the subset_data
    #         subset_data = self._add_label(subset_data, data)

    #         return subset_data

    def get(self, idx):
        """Initialize LMDB if it hasn't been initialized yet."""
        if self.env is None:
            self._init_db()

        with self.env.begin() as txn:
            serialized_data = txn.get(f"{idx}".encode())
            if serialized_data is None:
                return None
            data = pickle.loads(serialized_data)
            if self.transform:
                data = self.transform(data)

            subset_data = self._subset_graph(data)
            subset_data = self._add_label(subset_data, data)
            return subset_data

    def _init_db(self):
        """Initialize the LMDB environment."""
        self.env = lmdb.open(
            osp.join(self.processed_dir, "data.lmdb"),
            readonly=True,
            lock=False,
            readahead=False,
            meminit=False,
        )

    def len(self) -> int:
        if self.env is None:
            self._init_db()

        with self.env.begin() as txn:
            length = txn.stat()["entries"]

        # Must be closed for dataloader num_workers > 0
        self.close_lmdb()

        return length

    def close_lmdb(self):
        if self.env is not None:
            self.env.close()
            self.env = None


def main():
    # genome
    import os.path as osp

    from dotenv import load_dotenv

    from torchcell.datasets import OneHotGeneDataset
    from torchcell.graph import SCerevisiaeGraph

    load_dotenv()
    DATA_ROOT = os.getenv("DATA_ROOT")

    genome = SCerevisiaeGenome(
        data_root=osp.join(DATA_ROOT, "data/sgd/genome"), overwrite=True
    )
    graph = SCerevisiaeGraph(
        data_root=osp.join(DATA_ROOT, "data/sgd/genome"), genome=genome
    )
    dmf_dataset = DmfCostanzo2016Dataset(
        root=osp.join(DATA_ROOT, "data/scerevisiae/costanzo2016_1e5"),
        preprocess={"duplicate_resolution": "low_dmf_std"},
        # subset_n=100,
    )
    smf_dataset = SmfCostanzo2016Dataset(
        root=osp.join(DATA_ROOT, "data/scerevisiae/costanzo2016_smf"),
        preprocess={"duplicate_resolution": "low_std_both"},
        skip_process_file_exist_check=True,
    )
    gene_set = smf_dataset.gene_set.union(dmf_dataset.gene_set)
    #
    G = graph.G_go.copy()
    # Filtering
    G = filter_by_date(G, "2017-07-19")
    G = filter_go_IGI(G)
    G = filter_redundant_terms(G)
    G = filter_by_contained_genes(G, n=1, gene_set=gene_set)

    # replace graph
    graph.G_go = G

    # Instantiate the model
    dcell_model = DCell(go_graph=graph.G_go)

    one_hot = OneHotGeneDataset(root="data/scerevisiae/gene_one_hot", genome=genome)
    embeddings = one_hot

    # Experiments
    experiments = DmfCostanzo2016Dataset(
        preprocess={"duplicate_resolution": "low_dmf_std"},
        root=osp.join(DATA_ROOT, "data/scerevisiae/costanzo2016_1e3"),
        subset_n=1000,
    )

    cell_dataset = DCellDataset(
        root="data/scerevisiae/cell_1e3",
        genome=genome,
        graph=graph,
        embeddings=embeddings,
        experiments=experiments,
    )

    print(cell_dataset)
    print(cell_dataset.gene_set)
    print(cell_dataset[0])
    print(cell_dataset.wt)
    from torch_geometric.data import Batch

    batch = Batch.from_data_list([cell_dataset[0], cell_dataset[1]])
    dcell_subsystem_output = dcell_model(batch)
    dcell_linear = DCellLinear(dcell_model.subsystems, output_size=1)
    dcell_linear_output = dcell_linear(dcell_subsystem_output)
    print(dcell_linear_output)


if __name__ == "__main__":
    main()



# torchcell/dataset_readers/reader.py
# [[torchcell.dataset_readers.reader]]
# https://github.com/Mjvolk3/torchcell/tree/main/torchcell/dataset_readers/reader.py
# Test file: tests/torchcell/dataset_readers/test_reader.py

import os.path as osp
import pickle
import lmdb
import numpy as np
import json

from torchcell.datamodels import BaseGenotype, InterferenceGenotype
from torchcell.data.data import ExperimentReferenceIndex

class LmdbDatasetReader:
    def __init__(self, dataset_dir):
        self.dataset_dir = dataset_dir
        self.env = None
        self._experiment_reference_index = None
        self.db = None  # Add a db attribute
        self._init_db()
        self._load_experiment_reference_index()

    def _init_db(self):
        """Initialize the LMDB environment."""
        self.env = lmdb.open(
            osp.join(self.dataset_dir, "processed/data.lmdb"),
            readonly=True,
            lock=False,
            readahead=False,
            meminit=False,
            max_dbs=20
        )
        # Open the default database
        with self.env.begin(write=False) as txn:
            self.db = self.env.open_db(None, create=False)  # None refers to the default database


    def _load_experiment_reference_index(self):
        index_file_path = osp.join(
            self.dataset_dir, "preprocess/experiment_reference_index.json"
        )
        if osp.exists(index_file_path):
            with open(index_file_path, "r") as file:
                index_data = json.load(file)
                # Convert dictionaries to ExperimentReferenceIndex objects
                self._experiment_reference_index = [ExperimentReferenceIndex(**item) for item in index_data]

    @property
    def experiment_reference_index(self):
        return self._experiment_reference_index

    def save_experiment_reference_index(self, index_data):
        index_file_path = osp.join(
            self.dataset_dir, "preprocess/experiment_reference_index.json"
        )
        with open(index_file_path, "w") as file:
            json.dump(index_data, file, indent=4)

    def close_lmdb(self):
        if self.env is not None:
            self.env.close()
            self.env = None

    def __getitem__(self, idx):
        if isinstance(idx, (list, np.ndarray)):
            if isinstance(idx, list):
                idx = np.array(idx)
            if idx.dtype == np.bool_:
                idx = np.where(idx)[0]
            return [self.get_single_item(i) for i in idx]
        else:
            return self.get_single_item(idx)

    def get_single_item(self, idx):
        with self.env.begin() as txn:
            serialized_data = txn.get(f"{idx}".encode())
            if serialized_data is None:
                return None
            deserialized_data = pickle.loads(serialized_data)
            return deserialized_data

    def __len__(self):
        with self.env.begin() as txn:
            # Use the database handle
            length = txn.stat(db=self.db)["entries"]
        return length

    def __iter__(self):
        """Return an iterator over the dataset."""
        for idx in range(len(self)):
            yield self[idx]


    def __repr__(self):
        return f"{self.__class__.__name__}({len(self)})"



# torchcell/datasets/embedding.py
# [[torchcell.datasets.embedding]]
# https://github.com/Mjvolk3/torchcell/tree/main/torchcell/datasets/embedding.py
# Test file: torchcell/datasets/test_embedding.py

import os.path as osp
from abc import ABC, abstractmethod
from collections.abc import Callable
from typing import Optional

import torch
from torch_geometric.data import Data, InMemoryDataset
from tqdm import tqdm

from torchcell.sequence.genome.scerevisiae.s288c import SCerevisiaeGenome


class BaseEmbeddingDataset(InMemoryDataset, ABC):
    def __init__(
        self,
        root: str,
        # genome: SCerevisiaeGenome,  # If we include this we get ddp error
        model_name: str | None = None,
        transform: Callable | None = None,
        pre_transform: Callable | None = None,
    ):
        if model_name and model_name not in self.MODEL_TO_WINDOW:
            valid_model_names = ", ".join(self.MODEL_TO_WINDOW.keys())
            raise ValueError(
                f"Invalid model_name '{model_name}'."
                f"Valid options are: {valid_model_names}"
            )
        # BUG
        # self.genome = genome # If we include this we get ddp error
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.model_name = model_name
        super().__init__(root, transform, pre_transform)
        if self.model_name:
            self.data, self.slices = torch.load(
                self.processed_paths[0], map_location=self.device
            )
        else:
            self.data, self.slices = None, None

    @abstractmethod
    def initialize_model(self):
        pass

    @property
    def raw_file_names(self) -> list[str]:
        return []

    @property
    def processed_file_names(self) -> str:
        # if not self.model_name:
        # return "dummy_data.pt"
        return f"{self.model_name}.pt"

    def download(self):
        pass

    @abstractmethod
    def process(self):
        pass

    def get_data_list(self):
        return [data for data in self]

    def __getitem__(self, idx):
        if isinstance(idx, str):
            # Use _data instead of data to suppress warning. might be dangerous.
            if idx in self._data.id:
                index = self._data.id.index(idx)
                return Data(
                    id=self._data.id[index],
                    dna_windows={
                        key: value[index]
                        for key, value in self._data.dna_windows.items()
                    },
                    embeddings={
                        key: value[index : index + 1]  # Ensure 2D shape
                        for key, value in self._data.embeddings.items()
                    },
                )
            else:
                raise KeyError(f"Gene {idx} not found in the dataset.")
        else:
            return super().__getitem__(idx)

    def __radd__(self, other):
        # if 'other' is the default integer 0, return the current instance
        if isinstance(other, int) and other == 0:
            return self
        # Otherwise, just fall back to the normal add operation
        return self.__add__(other)

    def __add__(self, other):
        if isinstance(other, int) and other == 0:
            return self
        # Ensure the other object is of the same type
        if not isinstance(other, BaseEmbeddingDataset):
            raise ValueError("Can only add datasets of the same type.")

        combined_data_list = []

        # Create a dictionary from the current dataset for efficient lookup
        current_data_dict = {data_item.id: data_item for data_item in self}

        # Lists to store duplicate keys
        duplicate_dna_windows_keys = []
        duplicate_embeddings_keys = []

        # Combine the data from the other dataset
        for data_item in other:
            if data_item.id in current_data_dict:
                # Check for duplicate keys in dna_windows
                for key in data_item.dna_windows:
                    if key in current_data_dict[data_item.id].dna_windows:
                        duplicate_dna_windows_keys.append(key)
                    else:
                        # Merge the dna_windows dictionaries
                        current_data_dict[data_item.id].dna_windows[
                            key
                        ] = data_item.dna_windows[key]
                # Check for duplicate keys in embeddings
                for key in data_item.embeddings:
                    if key in current_data_dict[data_item.id].embeddings:
                        duplicate_embeddings_keys.append(key)
                    else:
                        # Merge the embeddings dictionaries
                        current_data_dict[data_item.id].embeddings[
                            key
                        ] = data_item.embeddings[key]
            else:
                combined_data_list.append(data_item)

        # If there are duplicates, raise an error
        if duplicate_dna_windows_keys:
            raise ValueError(
                "Duplicate keys found in dna_windows:"
                f"{', '.join(duplicate_dna_windows_keys)}"
            )
        if duplicate_embeddings_keys:
            raise ValueError(
                "Duplicate keys found in embeddings:"
                f"{', '.join(duplicate_embeddings_keys)}"
            )

        # Add the modified data items from the current dataset to the combined list
        combined_data_list.extend(current_data_dict.values())

        # Use collate to convert the combined list into the format InMemoryDataset expects
        data, slices = self.collate(combined_data_list)

        # Create a new dataset instance with the combined data
        combined_dataset = CombinedEmbedding(root=self.root, model_name=None)

        combined_dataset.data, combined_dataset.slices = data, slices

        return combined_dataset


# TODO, not sure if necessary but it renames dataset which is nice and clear,
# might not need ot inherit from BaseEmbeddingDataset since we have to implement
# initialize_model and process with pass
class CombinedEmbedding(BaseEmbeddingDataset):
    """
    This class represents the combined dataset which can
    be the result of combining two other datasets.
    """

    # No need to override the `__init__` method if it's going
    # to be the same as the parent class. However, if you need
    # any specialized initialization, you can define the method here.

    def initialize_model(self):
        """
        Implementation of abstract method.
        """
        # Provide a proper implementation if needed or just pass.
        pass

    def process(self):
        """
        Implementation of abstract method.
        """
        # Provide a proper implementation if needed or just pass.
        pass


if __name__ == "__main__":
    pass



from .act import act_register
from .constants import DNA_LLM_MAX_TOKEN_SIZE
from .dcell import DCell, DCellLinear, dcell_from_networkx
from .deep_set import DeepSet
from .deep_set_transformer import DeepSetTransformer
from .fungal_up_down_transformer import FungalUpDownTransformer
from .graph_attention import GraphAttention
from .graph_convolution import GraphConvolution
from .linear import SimpleLinearModel
from .mlp import Mlp
from .nucleotide_transformer import NucleotideTransformer

model_constants = ["DNA_LLM_MAX_TOKEN_SIZE"]

model_building_blocks = ["act_register"]

simple_models = ["Mlp"]

models = [
    "FungalUpDownTransformer",
    "NucleotideTransformer",
    "DeepSet",
    "DeepSetTransformer",
    "SimpleLinearModel",
    "GraphConvolution",
    "GraphAttention",
]

incidence_models = ["DCell", "DCellLinear"]

__all__ = (
    model_constants + simple_models + model_building_blocks + models + incidence_models
)



import math
import tracemalloc

import lightning as L
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import torch
import torch.nn as nn
from torch_geometric.data import Batch, Data
from torchmetrics import (
    MeanAbsoluteError,
    MeanSquaredError,
    MetricCollection,
    PearsonCorrCoef,
    SpearmanCorrCoef,
)
from tqdm import tqdm

import wandb
from torchcell.losses import DCellLoss, WeightedMSELoss
from torchcell.viz import fitness, genetic_interaction_score
import torchcell

style_file_path = osp.join(osp.dirname(torchcell.__file__), 'torchcell.mplstyle')
plt.style.use(style_file_path)

class DCellRegressionTask(L.LightningModule):
    """LightningModule for training models on graph-based regression datasets."""

    def __init__(
        self,
        models: dict[str, nn.Module],
        target: str,
        boxplot_every_n_epochs: int = 10,
        learning_rate: float = 1e-3,
        weight_decay: float = 1e-5,
        batch_size: int = None,
        train_wt_diff: bool = True,
        **kwargs,
    ):
        super().__init__()
        # models
        self.models = models

        for key, value in models.items():
            setattr(self, key, value)

        # target for training
        self.target = target

        # Lightning settings, doing this for WT embedding
        self.automatic_optimization = False

        self.x_name = "x"
        self.x_batch_name = "batch"

        self.loss = DCellLoss()

        # optimizer
        self.learning_rate = learning_rate
        self.weight_decay = weight_decay

        # batch_size
        self.batch_size = batch_size

        self.train_metrics = MetricCollection(
            {
                "RMSE": MeanSquaredError(squared=False),
                "MSE": MeanSquaredError(squared=True),
                "MAE": MeanAbsoluteError(),
            },
            prefix="train_",
        )
        self.val_metrics = self.train_metrics.clone(prefix="val_")
        self.test_metrics = self.train_metrics.clone(prefix="test_")

        # Separate attributes for Pearson and Spearman correlation coefficients
        self.pearson_corr = PearsonCorrCoef()
        self.spearman_corr = SpearmanCorrCoef()

        # Used in end for whisker plot
        self.boxplot_every_n_epochs = boxplot_every_n_epochs

        # wandb model artifact logging
        self.last_logged_best_step = None
        tracemalloc.start()

    def setup(self, stage=None):
        for model in self.models.values():
            model.to(self.device)
        self.true_values = torch.tensor([], dtype=torch.float32, device=self.device)
        self.predictions = torch.tensor([], dtype=torch.float32, device=self.device)

    def forward(self, batch):
        # Implement the forward pass
        dcell_subsystem_output = self.dcell(batch)
        dcell_linear_output = self.dcell_linear(dcell_subsystem_output)
        # if dcell_linear_output.size()[-1] == 1:
        #     dcell_linear_output = dcell_linear_output.squeeze(-1)
        return dcell_linear_output

    def on_train_start(self):
        # Calculate the model size (number of parameters)
        parameter_size = sum(p.numel() for p in self.parameters())
        # Log it using wandb
        self.log(
            "model/parameters_size", torch.tensor(parameter_size, dtype=torch.float32)
        )

    def training_step(self, batch, batch_idx):
        y_hat = self(batch)
        y = batch.fitness
        opt = self.optimizers()
        opt.zero_grad()
        loss = self.loss(y_hat, y, self.dcell.parameters())

        self.manual_backward(loss)  # error on this line
        opt.step()
        opt.zero_grad()
        # logging
        batch_size = batch.batch[-1].item() + 1
        # Flatten
        y_hat_root = y_hat["GO:ROOT"].squeeze(1)
        y_hat_stacked = torch.stack([v.squeeze() for v in y_hat.values()])
        y_hat_subsystems = y_hat_stacked.mean(0)
        # Log
        self.log("train_loss", loss, batch_size=batch_size, sync_dist=True)
        self.train_metrics(y_hat_subsystems, y)
        self.train_metrics(y_hat_root, y)
        # Logging the correlation coefficients
        self.log(
            "train_pearson_subsystems",
            self.pearson_corr(y_hat_subsystems, y),
            batch_size=batch_size,
            sync_dist=True,
        )
        self.log(
            "train_spearman_subsystems",
            self.spearman_corr(y_hat_subsystems, y),
            batch_size=batch_size,
            sync_dist=True,
        )
        self.log(
            "train_pearson_root",
            self.pearson_corr(y_hat_root, y),
            batch_size=batch_size,
            sync_dist=True,
        )
        self.log(
            "train_spearman_root",
            self.spearman_corr(y_hat_root, y),
            batch_size=batch_size,
            sync_dist=True,
        )
        return loss

    def on_train_epoch_end(self):
        self.log_dict(self.train_metrics.compute(), sync_dist=True)
        self.train_metrics.reset()

    def validation_step(self, batch, batch_idx):
        # Extract the batch vector
        y_hat = self(batch)
        y = batch.fitness
        loss = self.loss(y_hat, y, self.dcell.parameters())
        batch_size = batch.batch[-1].item() + 1
        self.log("val_loss", loss, batch_size=batch_size, sync_dist=True)
        # Flatten
        y_hat_root = y_hat["GO:ROOT"].squeeze(1)
        y_hat_stacked = torch.stack([v.squeeze() for v in y_hat.values()])
        y_hat_subsystems = y_hat_stacked.mean(0)
        # Log
        self.val_metrics(y_hat_subsystems, y)
        self.val_metrics(y_hat_root, y)
        # Logging the correlation coefficients
        self.log(
            "val_pearson_subsystems",
            self.pearson_corr(y_hat_subsystems, y),
            batch_size=batch_size,
            sync_dist=True,
        )
        self.log(
            "val_spearman_subsystems",
            self.spearman_corr(y_hat_subsystems, y),
            batch_size=batch_size,
            sync_dist=True,
        )
        self.log(
            "val_pearson",
            self.pearson_corr(y_hat_root, y),
            batch_size=batch_size,
            sync_dist=True,
        )
        self.log(
            "val_spearman",
            self.spearman_corr(y_hat_root, y),
            batch_size=batch_size,
            sync_dist=True,
        )
        self.true_values = torch.cat([self.true_values, y.detach()], dim=0)
        self.predictions = torch.cat(
            [self.predictions, y_hat_subsystems.detach()], dim=0
        )

    def on_validation_epoch_end(self):
        self.log_dict(self.val_metrics.compute(), sync_dist=True)
        self.val_metrics.reset()

        # Skip plotting during sanity check
        if self.trainer.sanity_checking or (
            self.current_epoch % self.boxplot_every_n_epochs != 0
        ):
            return

        if self.target == "fitness":
            fig = fitness.box_plot(self.true_values, self.predictions)
        elif self.target == "genetic_interaction_score":
            fig = genetic_interaction_score.box_plot(true_values, predictions)
        wandb.log({"binned_values_box_plot": wandb.Image(fig)})
        plt.close(fig)
        # Clear the stored values for the next epoch
        self.true_values = torch.tensor([], dtype=torch.float32, device=self.device)
        self.predictions = torch.tensor([], dtype=torch.float32, device=self.device)

        current_global_step = self.global_step
        # HACK
        # Get the current memory usage
        current, peak = tracemalloc.get_traced_memory()
        print("======")
        print(f"Current memory usage is {current / 10**6}MB; Peak was {peak / 10**6}MB")
        print("======")

        # Stop tracing memory allocations
        tracemalloc.stop()
        if (
            self.trainer.checkpoint_callback.best_model_path
            and current_global_step != self.last_logged_best_step
        ):
            # Save model as a W&B artifact
            artifact = wandb.Artifact(
                name=f"model-global_step-{current_global_step}",
                type="model",
                description=f"Model on validation epoch end step - {current_global_step}",
                metadata=dict(self.hparams),
            )
            artifact.add_file(self.trainer.checkpoint_callback.best_model_path)
            wandb.log_artifact(artifact)
            self.last_logged_best_step = (
                current_global_step  # update the last logged step
            )

    def test_step(self, batch, batch_idx):
        y_hat = self(batch)
        y = batch.fitness
        loss = self.loss(y_hat, y, self.dcell.parameters())
        batch_size = batch.batch[-1].item() + 1
        # Flatten
        y_hat_root = y_hat["GO:ROOT"].squeeze(1)
        y_hat_stacked = torch.stack([v.squeeze() for v in y_hat.values()])
        y_hat_subsystems = y_hat_stacked.mean(0)
        #
        self.log("test_loss", loss, batch_size=batch_size, sync_dist=True)
        self.test_metrics(y_hat_subsystems, y)
        self.test_metrics(y_hat_root, y)
        # Logging the correlation coefficients
        self.log(
            "test_pearson_subsystems",
            self.pearson_corr(y_hat_subsystems, y),
            batch_size=batch_size,
            sync_dist=True,
        )
        self.log(
            "test_spearman_subsystems",
            self.spearman_corr(y_hat_subsystems, y),
            batch_size=batch_size,
            sync_dist=True,
        )
        self.log(
            "test_pearson_root",
            self.pearson_corr(y_hat_root, y),
            batch_size=batch_size,
            sync_dist=True,
        )
        self.log(
            "test_spearman_root",
            self.spearman_corr(y_hat_root, y),
            batch_size=batch_size,
            sync_dist=True,
        )

    def on_test_epoch_end(self):
        self.log_dict(self.test_metrics.compute(), sync_dist=True)
        self.test_metrics.reset()

    def configure_optimizers(self):
        params = list(self.models["dcell"].parameters()) + list(
            self.models["dcell_linear"].parameters()
        )
        optimizer = torch.optim.Adam(
            params, lr=self.learning_rate, weight_decay=self.weight_decay
        )
        return optimizer



# torchcell/viz/fitness.py
# [[torchcell.viz.fitness]]
# https://github.com/Mjvolk3/torchcell/tree/main/torchcell/viz/fitness.py
# Test file: torchcell/viz/test_fitness.py

import matplotlib.pyplot as plt
import numpy as np
import torch


def box_plot(true_values: torch.tensor, predictions: torch.tensor) -> plt.Figure:
    # Define bins
    bins = [0.0, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0, 1.1, 1.2, float("inf")]

    # font name
    font_name = "DejaVu Sans"

    # Bin predictions and collect corresponding true values
    binned_true_values = []

    for i in range(len(bins) - 1):
        mask = (predictions >= bins[i]) & (predictions < bins[i + 1])
        binned_values = true_values[mask].cpu().numpy()
        binned_true_values.append(binned_values)

    # Create a box plot using matplotlib
    # width / height
    aspect_ratio = 1.18
    height = 6
    width = height * aspect_ratio

    fig, ax = plt.subplots(figsize=(width, height), dpi=140)

    # Equally spaced box positions
    box_positions = [i + 0.5 for i in range(len(bins) - 1)]

    # Compute tick values
    xticks = [str(round(bin_val, 2)) for bin_val in bins[:-1]]
    # Add 'inf' as the last tick label
    xticks.append("Inf")
    # Tick positions
    tick_positions = [i for i in range(len(bins))]

    # Plot the vertical grey lines
    # Set zorder to 0 to be behind boxes
    for pos in tick_positions:
        ax.axvline(x=pos, color="#838383", linewidth=0.8, zorder=0)

    # set the spine - outside box
    for spine in ax.spines.values():
        spine.set_linewidth(1.4)

    # Set zorder to 1 to be above the horizontal line
    # Adjust for box width
    boxplots = ax.boxplot(
        binned_true_values,
        patch_artist=True,
        vert=True,
        widths=0.98,
        positions=box_positions,
        showfliers=False,
        capwidths=0,
        zorder=1,
    )

    # Apply coloring to the boxes, whiskers, and medians
    for patch in boxplots["boxes"]:
        patch.set_facecolor("#F6A9A3")
        patch.set_edgecolor("#D86B2B")
        # Increase outline width
        patch.set_linewidth(2.2)
    for whisker in boxplots["whiskers"]:
        whisker.set_color("#D86B2B")
        # Increase whisker width
        whisker.set_linewidth(2.0)
    for median in boxplots["medians"]:
        median.set_color("#D86B2B")
        # Increase median width
        median.set_linewidth(4.0)
        # Get current x-values (endpoints) of the median line
        x = median.get_xdata()
        # Adjust the endpoints to reduce the width
        width_reduction = 0.05
        x[0] += width_reduction
        x[1] -= width_reduction
        # Set the modified x-values back to the median line
        median.set_xdata(x)

    # Add a black horizontal line at y=1.0
    # Set zorder to 0 to be behind boxes
    ax.axhline(y=1, color="black", linewidth=1.4, zorder=2)

    # Add a black vertical line at x=7.0 (1.0 on the tick label)
    # Set zorder to 2 to be above boxes
    ax.axvline(x=7.0, color="black", linewidth=1.4, zorder=2)

    # Add "(WT)" labels to x and y axes
    ax.text(-0.96, 0.995, "(WT)", fontsize=17.0, va="center", ha="right")
    ax.text(7.48, -0.08, "(WT)", fontsize=17.0, va="center", ha="right")

    # Set tick labels
    ax.set_xticks(tick_positions)
    ax.set_xticklabels(xticks, ha="center")

    # Adjust x and y label positions
    ax.set_xlabel("Predicted growth", labelpad=34, size=17.0, fontname=font_name)
    ax.set_ylabel("Measured growth", labelpad=29, size=17.0, fontname=font_name)

    # Set y-axis limits and ticks
    ax.set_ylim(0.05, 1.25)
    # Show only till 1.2 with spacing of 0.1
    ax.set_yticks(np.arange(0.1, 1.3, 0.1))

    # Set the size of xticks and yticks
    ax.tick_params(axis="x", length=4, width=0, labelsize=16.0)
    ax.tick_params(axis="y", length=7, width=1.4, labelsize=16.0)

    # For x-ticks
    for label in ax.get_xticklabels():
        label.set_fontname(font_name)

    # For y-ticks
    for label in ax.get_yticklabels():
        label.set_fontname(font_name)

    plt.tight_layout()

    return fig


def main():
    # import matplotlib.font_manager as fm

    # fonts = fm.findSystemFonts()
    # for font in fonts:
    #     print(font)
    predictions = torch.load("predictions.pt")
    true_values = torch.load("true_values.pt")
    box_plot(true_values, predictions)
    plt.show()


if __name__ == "__main__":
    main()



# torchcell/viz/genetic_interaction_score.py
# [[torchcell.viz.genetic_interaction_score]]
# https://github.com/Mjvolk3/torchcell/tree/main/torchcell/viz/genetic_interaction_score.py
# Test file: torchcell/viz/test_genetic_interaction_score.py

import matplotlib.pyplot as plt
import numpy as np
import torch


def box_plot(true_values: torch.tensor, predictions: torch.tensor) -> plt.Figure:
    # Define bins for the second image
    bins = [
        -float("inf"),
        -0.40,
        -0.32,
        -0.24,
        -0.16,
        -0.08,
        0.00,
        0.08,
        0.16,
        0.24,
        float("inf"),
    ]

    # font name
    font_name = "DejaVu Sans"

    # Bin predictions and collect corresponding true values
    binned_true_values = []

    for i in range(len(bins) - 1):
        mask = (predictions >= bins[i]) & (predictions < bins[i + 1])
        binned_values = true_values[mask].cpu().numpy()
        binned_true_values.append(binned_values)

    # Create a box plot using matplotlib
    # width / height
    aspect_ratio = 1.18
    height = 6
    width = height * aspect_ratio

    fig, ax = plt.subplots(figsize=(width, height), dpi=140)

    # Equally spaced box positions
    box_positions = [i + 0.5 for i in range(len(bins) - 1)]

    # Compute tick values with three decimal places
    xticks = [f"{bin_val:.2f}" for bin_val in bins[:-1]]
    # Add 'inf' as the first and last tick label
    xticks[0] = "-Inf"
    xticks.append("Inf")

    # Tick positions
    tick_positions = [i for i in range(len(bins))]

    # Plot the vertical grey lines
    for pos in tick_positions:
        ax.axvline(x=pos, color="#838383", linewidth=0.8, zorder=0)

    # Set spines
    for spine in ax.spines.values():
        spine.set_linewidth(1.4)

    # Boxplots
    boxplots = ax.boxplot(
        binned_true_values,
        patch_artist=True,
        vert=True,
        widths=0.98,
        positions=box_positions,
        showfliers=False,
        capwidths=0,
        zorder=1,
    )

    # Apply coloring
    for patch in boxplots["boxes"]:
        patch.set_facecolor("#F6A9A3")
        patch.set_edgecolor("#D86B2B")
        patch.set_linewidth(2.2)
    for whisker in boxplots["whiskers"]:
        whisker.set_color("#D86B2B")
        whisker.set_linewidth(2.0)
    for median in boxplots["medians"]:
        median.set_color("#D86B2B")
        median.set_linewidth(4.0)
        x = median.get_xdata()
        width_reduction = 0.05
        x[0] += width_reduction
        x[1] -= width_reduction
        median.set_xdata(x)

    # Add a black horizontal line at y=0
    ax.axhline(y=0, color="black", linewidth=1.4, zorder=2)
    # Add a vertical black line at x=0
    x_position_for_zero_bin = bins.index(0.00)
    ax.axvline(x=x_position_for_zero_bin, color="black", linewidth=1.4, zorder=2)
    # Set tick labels
    ax.set_xticks(tick_positions)
    ax.set_xticklabels(xticks, ha="center", rotation=45, fontsize=16.0)
    # After setting your x-ticks
    labels = [item.get_text() for item in ax.get_xticklabels()]
    formatted_labels = [
        label.replace("-", "–") if "-" in label else label for label in labels
    ]
    ax.set_xticklabels(formatted_labels)

    # Adjust x and y label positions
    ax.set_xlabel(
        "Predicted genetic interaction", labelpad=8, size=17.0, fontname=font_name
    )
    ax.set_ylabel(
        "Measured genetic interaction", labelpad=8, size=17.0, fontname=font_name
    )

    # Set y-axis limits and ticks
    y_ticks = [-0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4]
    # Leave half space above max and below min
    y_min, y_max = (min(y_ticks) - 0.05, max(y_ticks) + 0.05)
    ax.set_ylim(y_min, y_max)
    ax.set_yticks(y_ticks)

    # Set tick size
    ax.tick_params(axis="x", length=4, width=0, labelsize=16.0)
    ax.tick_params(axis="y", length=7, width=1.6, labelsize=16.0)

    # Font adjustments
    for label in ax.get_xticklabels():
        label.set_fontname(font_name)
    for label in ax.get_yticklabels():
        label.set_fontname(font_name)

    plt.tight_layout()

    return fig



# torchcell/datasets/fungal_up_down_transformer.py
# [[torchcell.datasets.fungal_up_down_transformer]]
# https://github.com/Mjvolk3/torchcell/tree/main/torchcell/datasets/fungal_up_down_transformer.py
# Test file: torchcell/datasets/test_fungal_up_down_transformer.py

import os
from collections.abc import Callable
from typing import Optional

import torch
from pydantic import validator
from torch_geometric.data import Data
from tqdm import tqdm

from torchcell.datamodels import ModelStrictArbitrary
from torchcell.datasets.embedding import BaseEmbeddingDataset
from torchcell.models.fungal_up_down_transformer import (  # adjusted import
    FungalUpDownTransformer,
)
from torchcell.sequence import GeneSet, ParsedGenome
from torchcell.sequence.genome.scerevisiae.s288c import SCerevisiaeGenome


class FungalUpDownTransformerDataset(BaseEmbeddingDataset):
    MODEL_TO_WINDOW = {
        "species_downstream": ("window_three_prime", 300, True, True),
        "species_upstream": ("window_five_prime", 1003, True, True),
    }

    def __init__(
        self,
        root: str,
        genome: SCerevisiaeGenome,
        model_name: str | None = None,
        transform: Callable | None = None,
        pre_transform: Callable | None = None,
    ):
        self.genome = genome
        self.model_name = model_name
        # BUG I just moved this here to recompute the data but we don't want to do this when training models
        self.transformer = self.initialize_model()
        super().__init__(root, self.model_name, transform, pre_transform)
        # convert genome to parsed genome after process, so have potential issue
        # with sqlite database
        # TODO try without parsed_genome on ddp to see if issue was
        # BaseEmbeddingDataset previously taking genome as a parameter
        self.genome = self.parse_genome(genome)
        del genome

        if self.model_name:
            if not os.path.exists(self.processed_paths[0]):
                self.transformer = self.initialize_model()
                self.process()
            self.data, self.slices = torch.load(self.processed_paths[0])

    @staticmethod
    def parse_genome(genome) -> ParsedGenome:
        # BUG we have to do this black magic because when you merge datasets with +
        # the genome is None
        if genome is None:
            return None
        else:
            data = {}
            data["gene_set"] = genome.gene_set
            return ParsedGenome(**data)

    def initialize_model(self) -> FungalUpDownTransformer | None:
        if self.model_name:
            split_name = self.model_name.split("_")
            if "downstream" in split_name and "species" in split_name:
                model_name = "downstream_species_lm"
            elif "upstream" in split_name and "species" in split_name:
                model_name = "upstream_species_lm"
            assert (
                model_name in FungalUpDownTransformer.VALID_MODEL_NAMES
            ), f"{model_name} not in valid model names."
            return FungalUpDownTransformer(model_name)
        return None

    def process(self):
        if not self.model_name:
            return

        data_list = []
        (
            window_method,
            window_size,
            include_cds_codon,
            allow_undersize,
        ) = self.MODEL_TO_WINDOW[self.model_name]
        for gene_id in tqdm(self.genome.gene_set):
            sequence = self.genome[gene_id]
            dna_selection = getattr(sequence, window_method)(
                window_size, include_cds_codon, allow_undersize=allow_undersize
            )
            embeddings = self.transformer.embed(
                [dna_selection.seq], mean_embedding=True
            )

            dna_window_dict = {self.model_name: dna_selection}

            data = Data(id=gene_id, dna_windows=dna_window_dict)
            data.embeddings = {self.model_name: embeddings}
            data_list.append(data)

        if self.pre_transform:
            data_list = [self.pre_transform(data) for data in data_list]

        torch.save(self.collate(data_list), self.processed_paths[0])


if __name__ == "__main__":
    genome = SCerevisiaeGenome()

    # Adjust the model names accordingly.
    model_names = ["species_downstream", "species_upstream"]
    # model_names = ["species_upstream"]

    datasets = []
    for model_name in model_names:
        dataset = FungalUpDownTransformerDataset(
            root="data/scerevisiae/fungal_up_down_embed",
            genome=genome,
            model_name=model_name,
        )
        datasets.append(dataset)
        print(f"Dataset for {model_name}: {dataset}")
    print()
    # seq_embeddings = datasets[0] + datasets[1]
    # some_data = datasets[0][genome.gene_set[42]]
    # print(some_data)
    # some_data = datasets[1][genome.gene_set[42]]
    # print(some_data)



from cliff.command import Command
import argparse
import subprocess
import os
import os.path as osp


class BuildCommand(Command):
    "A simple command that builds something."

    def get_parser(self, prog_name):
        parser = super(BuildCommand, self).get_parser(prog_name)
        parser.add_argument(
            "-m",
            "--mode",
            choices=["regular", "fresh"],
            default="regular",
            help="Build mode (default: %(default)s)",
        )
        return parser

    def take_action(self, parsed_args):
        script_path = osp.join(os.getcwd(), "database", "build")
        script = (
            "build_linux-arm.sh"
            if parsed_args.mode == "regular"
            else "build_image_fresh_linux-arm.sh"
        )
        full_script_path = osp.join(script_path, script)

        print(
            f"{parsed_args.mode.capitalize()} mode: Executing script {full_script_path}"
        )

        # Ensure the script is executable
        os.chmod(full_script_path, 0o755)

        # Start the process
        process = subprocess.Popen(
            [full_script_path],
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            shell=True,
            text=True,
        )

        # Stream the output
        while True:
            output = process.stdout.readline()
            if output == "" and process.poll() is not None:
                break
            if output:
                print(output.strip())
        rc = process.poll()
        return rc



# torchcell/datasets/experiment.py
# [[torchcell.datasets.experiment]]
# https://github.com/Mjvolk3/torchcell/tree/main/torchcell/datasets/experiment.py
# Test file: torchcell/datasets/test_experiment.py

import json
import logging
import pickle
from collections.abc import Callable

import lmdb
from tqdm import tqdm

from torchcell.data import Dataset
from torchcell.datamodels import ModelStrictArbitrary
from torchcell.sequence import GeneSet

log = logging.getLogger(__name__)


# IDEA right now experiments: list[Dataset] but the Dataset[0]: Data
# will need to adhere to the same schema
# TODO unify Costanzo2016
class MergedExperiment(Dataset):
    def __init__(
        self,
        root: str = "data/scerevisiae/merged_experiment",
        experiments: list[Dataset] = None,
        preprocess: dict | None = None,
        skip_process_file_exist_check: bool = False,
        transform: Callable | None = None,
        pre_transform: Callable | None = None,
    ):
        self.experiments = experiments
        self.preprocess = preprocess
        self._skip_process_file_exist = skip_process_file_exist_check
        self.preprocess = preprocess
        self.preprocess_dir = osp.join(root, "preprocess")
        self._length = None
        self._gene_set = None
        self._df = None
        # Check for existing preprocess config
        existing_config = self.load_preprocess_config()
        if existing_config is not None:
            if existing_config != self.preprocess:
                raise ValueError(
                    "New preprocess does not match existing config."
                    "Delete the processed and process dir for a new Dataset."
                    "Or define a new root."
                )
        # Check for existing preprocess config
        super().__init__(root, transform, pre_transform)
        self.env = None

    @property
    def skip_process_file_exist(self):
        return self._skip_process_file_exist

    @property
    def raw_file_names(self) -> list[str]:
        return ["strain_ids_and_single_mutant_fitness.xlsx"]

    @property
    def processed_file_names(self) -> list[str]:
        return "data.lmdb"

    def _init_db(self):
        """Initialize the LMDB environment."""
        self.env = lmdb.open(
            osp.join(self.processed_dir, "data.lmdb"),
            readonly=True,
            lock=False,
            readahead=False,
            meminit=False,
        )

    def close_lmdb(self):
        if self.env is not None:
            self.env.close()
            self.env = None

    @property
    def wt(self):
        # TODO implement this, must be combination of all experiments
        pass

    # Merge Operations - This is HACK y
    def get_genotype_key(self, data_item):
        return data_item["genotype"]

    def get_environment_key(self, data_item):
        return data_item["phenotype"]["environment"]

    def get_phenotype_key(self, data_item):
        return data_item["phenotype"]["observation"]

    def contains_key(self, key, data_list):
        return key in [self.get_genotype_key(data_item) for data_item in data_list]

    def contains_environment_key(self, genotype_key, environment_key, data_list):
        return environment_key in [
            self.get_environment_key(data_item)
            for data_item in data_list
            if self.get_genotype_key(data_item) == genotype_key
        ]

    @property
    def experiment_indices(self):
        indices_path = osp.join(self.processed_dir, "experiment_indices.json")

        if osp.exists(indices_path):
            with open(indices_path) as f:
                return json.load(f)

        # If no cached indices found, log a warning and return an empty dict
        log.warning("No cached experiment indices found!")
        return {}

    def process(self):
        combined_data_list = []
        source_dataset_list = []  # New list to track the source dataset of each item

        # Extract the desired temperature from the preprocess dictionary, if specified
        desired_temperature = None
        if self.preprocess and "temperature" in self.preprocess:
            desired_temperature = self.preprocess["temperature"]

        for dataset in self.experiments:
            for data_item in dataset:
                # If a temperature preprocessing condition exists,
                # skip items not matching it
                if desired_temperature is not None:
                    current_temp = data_item["phenotype"]["environment"]["temperature"]
                    if current_temp != desired_temperature:
                        continue  # Skip this item

                genotype_key = self.get_genotype_key(data_item)
                environment_key = self.get_environment_key(data_item)
                phenotype_key = self.get_phenotype_key(data_item)

                if not self.contains_key(genotype_key, combined_data_list):
                    combined_data_list.append(data_item)
                    source_dataset_list.append(dataset.__class__.__name__)
                else:
                    if not self.contains_environment_key(
                        genotype_key, environment_key, combined_data_list
                    ):
                        combined_data_list.append(data_item)
                        source_dataset_list.append(dataset.__class__.__name__)
                    else:
                        # This checks if the phenotype key exists for the
                        # given genotype and environment.
                        # However, the function "contains_phenotype_key" is not defined
                        # in your provided code.
                        # You might want to implement it similarly to
                        # "contains_environment_key".
                        if not self.contains_phenotype_key(
                            genotype_key,
                            environment_key,
                            phenotype_key,
                            combined_data_list,
                        ):
                            combined_data_list.append(data_item)
                            source_dataset_list.append(dataset.__class__.__name__)
                        else:
                            log.warning(
                                f"Data item with genotype key: {genotype_key}, "
                                f"environment key: {environment_key}, and "
                                f"phenotype key: {phenotype_key} already exists in the merged dataset."
                            )

        # Initialize LMDB environment
        log.info("lmdb begin")
        env = lmdb.open(osp.join(self.processed_dir, "data.lmdb"), map_size=int(1e12))

        with env.begin(write=True) as txn:
            # Iterate through each data item
            for idx, item in tqdm(enumerate(combined_data_list)):
                serialized_data = pickle.dumps(item)
                txn.put(f"{idx}".encode(), serialized_data)

        # cache gene property
        self.gene_set = self.compute_gene_set(combined_data_list)

        # Now compute the experiment indices using source_dataset_list
        experiment_indices = {
            dataset.__class__.__name__: {"size": 0, "indices": []}
            for dataset in self.experiments
        }
        for idx, dataset_name in enumerate(source_dataset_list):
            experiment_indices[dataset_name]["indices"].append(idx)
            experiment_indices[dataset_name]["size"] += 1

        # Save the cached indices as a JSON file
        with open(osp.join(self.processed_dir, "experiment_indices.json"), "w") as f:
            json.dump(experiment_indices, f, indent=4)

    def get(self, idx):
        """Initialize LMDB if it hasn't been initialized yet."""
        if self.env is None:
            self._init_db()

        with self.env.begin() as txn:
            serialized_data = txn.get(f"{idx}".encode())
            if serialized_data is None:
                return None
            data = pickle.loads(serialized_data)
            if self.transform:
                data = self.transform(data)
            return data

    # New method to save preprocess configuration to a JSON file
    def save_preprocess_config(self, preprocess):
        if not osp.exists(self.preprocess_dir):
            os.makedirs(self.preprocess_dir)
        with open(osp.join(self.preprocess_dir, "preprocess_config.json"), "w") as f:
            json.dump(preprocess, f)

    # New method to load existing preprocess configuration
    def load_preprocess_config(self):
        config_path = osp.join(self.preprocess_dir, "preprocess_config.json")

        if osp.exists(config_path):
            with open(config_path) as f:
                config = json.load(f)
            return config
        else:
            return None

    def len(self) -> int:
        if self.env is None:
            self._init_db()

        with self.env.begin() as txn:
            length = txn.stat()["entries"]

        # Must be closed for dataloader num_workers > 0
        self.close_lmdb()

        return length

    @staticmethod
    def compute_gene_set(data_list):
        computed_gene_set = GeneSet()
        for data in data_list:
            for genotype in data.genotype:
                computed_gene_set.add(genotype["id"])
        return computed_gene_set

    # Reading from JSON and setting it to self._gene_set
    @property
    def gene_set(self):
        try:
            if osp.exists(osp.join(self.preprocess_dir, "gene_set.json")):
                with open(osp.join(self.preprocess_dir, "gene_set.json")) as f:
                    self._gene_set = GeneSet(json.load(f))
            elif self._gene_set is None:
                raise ValueError(
                    "gene_set not written during process. "
                    "Please call compute_gene_set in process."
                )
            return self._gene_set
        # CHECK can probably remove this
        except json.JSONDecodeError:
            raise ValueError("Invalid or empty JSON file found.")

    @gene_set.setter
    def gene_set(self, value):
        if not value:
            raise ValueError("Cannot set an empty or None value for gene_set")
        os.makedirs(self.preprocess_dir, exist_ok=True)
        with open(osp.join(self.preprocess_dir, "gene_set.json"), "w") as f:
            json.dump(list(sorted(value)), f, indent=0)
        self._gene_set = value

    def __repr__(self):
        return f"{self.__class__.__name__}({len(self)})"


if __name__ == "__main__":
    import os
    import os.path as osp

    from dotenv import load_dotenv

    from torchcell.datasets.scerevisiae import (
        DmfCostanzo2016Dataset,
        SmfCostanzo2016Dataset,
    )

    load_dotenv()
    DATA_ROOT = os.getenv("DATA_ROOT")

    dmf_dataset = DmfCostanzo2016Dataset(
        root=osp.join(DATA_ROOT, "data/scerevisiae/costanzo2016_1e2"),
        preprocess={"duplicate_resolution": "low_dmf_std"},
        subset_n=100,
    )
    print(dmf_dataset)
    print(dmf_dataset[0])
    print(len(dmf_dataset.gene_set))

    smf_dataset = SmfCostanzo2016Dataset(
        root=osp.join(DATA_ROOT, "data/scerevisiae/costanzo2016_smf"),
        preprocess={"duplicate_resolution": "low_std_both"},
        skip_process_file_exist_check=True,
        subset_n=100,
    )
    print(smf_dataset)
    print(smf_dataset[0])
    print(len(smf_dataset.gene_set))

    # TODO implement this
    fitness = MergedExperiment(
        root=osp.join(DATA_ROOT, "data/scerevisiae/smf_dmf"),
        experiments=[smf_dataset, dmf_dataset],
        preprocess={"temperature": 30},
    )
    print(fitness)
    print(fitness[0])
    print(fitness.experiment_indices)
    print(fitness.gene_set)



# torchcell/losses/dcell.py
# [[torchcell.losses.dcell]]
# https://github.com/Mjvolk3/torchcell/tree/main/torchcell/losses/dcell.py
# Test file: torchcell/losses/test_dcell.py

import torch
import torch.nn as nn


class DCellLoss(nn.Module):
    def __init__(self, alpha=0.3):
        super().__init__()
        self.alpha = alpha
        self.criterion = nn.MSELoss()

    def forward(self, outputs, target, weights):
        # Assuming 'GO:ROOT' is the root subsystem
        root_output = outputs["GO:ROOT"]
        root_loss = self.criterion(root_output.squeeze(-1), target)

        # Loss for non-root subsystems
        non_root_loss = sum(
            self.criterion(outputs[t].squeeze(-1), target)
            for t in outputs
            if t != "GO:ROOT"
        )

        # Total loss
        total_loss = root_loss + self.alpha * non_root_loss
        return total_loss



import lightning as L
import torch
import torch.nn as nn

from torchcell.datamodules import CellDataModule


class SimpleModel(L.LightningModule):
    def __init__(self):
        super().__init__()
        self.linear = nn.Linear(
            in_features=..., out_features=...
        )  # Define your model's architecture here

    def forward(self, x):
        return self.linear(x)

    def training_step(self, batch, batch_idx):
        # Define the training loop
        x, y = batch
        y_pred = self(x)
        loss = nn.MSELoss()(y_pred, y)  # Example loss for regression task
        return loss

    def validation_step(self, batch, batch_idx):
        x, y = batch
        y_pred = self(x)
        val_loss = nn.MSELoss()(y_pred, y)
        self.log("val_loss", val_loss)

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(self.parameters(), lr=1e-3)
        return optimizer


if __name__ == "__main__":
    data_module = CellDataModule(batch_size=64)
    model = SimpleModel()

    trainer = L.Trainer(max_epochs=10, gpus=1 if torch.cuda.is_available() else 0)
    trainer.fit(model, datamodule=data_module)



# torchcell/datasets/pronto_ontology.py
# [[torchcell.datasets.pronto_ontology]]
# https://github.com/Mjvolk3/torchcell/tree/main/torchcell/datasets/pronto_ontology.py
# Test file: torchcell/datasets/test_pronto_ontology.py

from collections import Counter

import matplotlib.pyplot as plt
import networkx as nx

# # Draw the graph
# pos = nx.spring_layout(G)
# labels = nx.get_node_attributes(G, 'label')
# nx.draw(G, pos, labels=labels, with_labels=True, node_size=3000, node_color='skyblue', font_size=10, font_weight='bold')
# plt.show()
import pronto
from nxontology.imports import multidigraph_to_digraph, pronto_to_multidigraph
from nxontology.viz import create_similarity_graphviz
from pronto import Definition, Ontology

# # https://obofoundry.org/ontology/mco.html
# url = "https://raw.githubusercontent.com/microbial-conditions-ontology/microbial-conditions-ontology/master/mco.owl"
# env_pronto = Ontology(handle=url)
# print(env_pronto.__dict__)
# print(50 * "=")
# print(len(env_pronto.terms()))
# [print(i) for i in list(env_pronto.terms())]
# print(50 * "=")
# [print(i) for i in list(env_pronto.relationships())]
# print(len(env_pronto.relationships()))

# # if __name__ == "__main__":
# #     Plotting the graph
# #     env_multidigraph = pronto_to_multidigraph(env_pronto)
# #     pos = nx.spring_layout(env_multidigraph)
# #     plt.figure(figsize=(10, 10))
# #     nx.draw(
# #         env_multidigraph, pos, with_labels=True, node_size=1000, node_color="skyblue"
# #     )
# #     plt.title("env_multidigraph")
# #     plt.show()
# #     pass

###############


# # Create a directed graph
# G = nx.DiGraph()

# # Add nodes for each term in the ontology
# for term in ontology:
#     G.add_node(term.id, label=term.name)

# # Add edges for each relationship in the ontology
# for term in ontology:
#     for rel, parent in term.relationships.items():
#         G.add_edge(term.id, parent.id, label=rel.id)


if __name__ == "__main__":
    # Create a new ontology
    ontology = pronto.Ontology()

    # Add terms to the ontology
    term1 = ontology.create_term("GO:0000001")
    term1.name = "mitochondrion inheritance"
    term1.definition = Definition(
        "The distribution of mitochondria, including the mitochondrial genome, into daughter cells after mitosis or meiosis, mediated by interactions between mitochondria and the cytoskeleton.",
        [],
    )

    term2 = ontology.create_term("GO:0000002")
    term2.name = "mitochondrial genome maintenance"
    term2.definition = Definition(
        "The maintenance of the structure and integrity of the mitochondrial genome; includes replication and segregation of the mitochondrial chromosome.",
        [],
    )

    # Use the predefined "is_a" relationship
    relationship = ontology.get_relationship("is_a")
    term2.relationships[relationship] = {term1}

    print(ontology)

    # Plotting the graph
    onto_multidigraph = pronto_to_multidigraph(ontology)
    pos = nx.spring_layout(onto_multidigraph)
    plt.figure(figsize=(10, 10))
    nx.draw(
        onto_multidigraph, pos, with_labels=True, node_size=1000, node_color="skyblue"
    )
    plt.title("onto_multidigraph")
    plt.show()



import json

import torch
from sklearn.linear_model import LinearRegression
from tqdm import tqdm
from transformers import AutoModelForMaskedLM, AutoTokenizer

from torchcell.data_priorsequence import SCerevisiaeGenome
from torchcell.datasets.scerevisiae import Baryshnikovna2010Dataset
from torchcell.models import DeepSet
from torchcell.sequence import roman_to_int

# Genome and dataset
genome = SCerevisiaeGenome()
dataset = Baryshnikovna2010Dataset()

# find overlap in datasetset and genome
dataset_ids = [id[0].split("_")[0] for id in dataset.genotype["ids"]]
gene_intersection = list(set(genome.gene_list).intersection(set(dataset_ids)))
print(f"len(gene_intersection): {len(gene_intersection)}")
# Get sequence data
data = {}
for i, gene in tqdm(enumerate(gene_intersection)):
    gene_feature = genome[gene]
    chr = roman_to_int((gene_feature.chrom).split("chr")[-1])
    start = gene_feature.start
    stop = gene_feature.stop
    strand = gene_feature.strand
    sequence = genome.get_seq(chr, start, stop, strand).seq
    if len(sequence) > 5994:
        print(f"skipping gene: {gene}, length: {len(sequence)}")
        continue
    # build data
    gene_data = {
        "index": i,
        "gene": gene,
        "sequence": sequence,
        "fitness": float(
            [
                dataset[i].phenotype["fitness"]
                for i in range(len(dataset))
                if dataset[i].genotype["ids"][0].split("_")[0] == gene
            ][0]
        ),
    }
    data[gene] = gene_data  # Use the gene as the key for each gene_data dictionary

# Dump the data dictionary to a JSON file
with open("data/bary_data.json", "w") as f:
    json.dump(data, f)


def main():
    print()


if __name__ == "__main__":
    main()



from .dataset import Dataset
from .cell_dataset import compute_experiment_reference_index
from .experiment_dataset import ExperimentDataset
from .experiment_dataset import post_process

datasets = ["Dataset", "ExperimentDataset"]
cell_dataset_functions = ["compute_experiment_reference_index"]

__all__ = datasets + cell_dataset_functions



# torchcell/knowledge_graphs/create_pypy_scerevisiae_kg
# [[torchcell.knowledge_graphs.create_pypy_scerevisiae_kg]]
# https://github.com/Mjvolk3/torchcell/tree/main/torchcell/knowledge_graphs/create_pypy_scerevisiae_kg
# Test file: tests/torchcell/knowledge_graphs/test_create_pypy_scerevisiae_kg.py


from biocypher import BioCypher
from torchcell.pypy_adapters import (
    SmfCostanzo2016Adapter,
    DmfCostanzo2016Adapter,
    SmfKuzmin2018Adapter,
    DmfKuzmin2018Adapter,
    TmfKuzmin2018Adapter,
)
from torchcell.dataset_readers import LmdbDatasetReader

import logging
from datetime import datetime

# Configure logging
logging.basicConfig(level=logging.INFO, filename="biocypher_warnings.log")
logging.captureWarnings(True)


def main():
    # logger.info(f"Started at {datetime.now()}") but use logging
    bc = BioCypher()

    # Ordered adapters from smallest to largest
    adapters = [
        DmfCostanzo2016Adapter(
            dataset=LmdbDatasetReader("data/torchcell/dmf_costanzo2016")
        ),
        SmfCostanzo2016Adapter(
            dataset=LmdbDatasetReader("data/torchcell/smf_costanzo2016")
        ),
        SmfKuzmin2018Adapter(
            dataset=LmdbDatasetReader("data/torchcell/smf_kuzmin2018")
        ),
        DmfKuzmin2018Adapter(
            dataset=LmdbDatasetReader("data/torchcell/dmf_kuzmin2018")
        ),
        TmfKuzmin2018Adapter(
            dataset=LmdbDatasetReader("data/torchcell/tmf_kuzmin2018")
        ),
    ]

    for adapter in adapters:
        bc.write_nodes(adapter.get_nodes())
        bc.write_edges(adapter.get_edges())

    # Write admin import statement and schema information (for biochatter)
    bc.write_import_call()
    bc.write_schema_info(as_node=True)

    # Print summary
    bc.summary()


if __name__ == "__main__":
    main()
